#include "config.h"

/*------------------------------------------------------------------------*/

#include <stdio.h>
#include "nanosat.h"

/*------------------------------------------------------------------------*/

#include <assert.h>
#include <stdlib.h>
#include <strings.h>
#include <ctype.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <stdarg.h>
#include <limits.h>

/*------------------------------------------------------------------------*/
/* Calculate signatures.
 *
#define SIG
 */

/*------------------------------------------------------------------------*/
/* Maximal number of variables accepted.
 */
#define MAXVARS (1 << 20)	/* roughly 4 million */

/*------------------------------------------------------------------------*/
/* Minimum size of a chunk of memory allocated.
 */
#define MIN_CHUNK_SIZE ((1 << 12) - 32)

/*------------------------------------------------------------------------*/
/* This is the maximum number of 'scale_down_variable_activities' before we
 * have to reinitialize the 'ld_activity_scale'.  For testing and debugging
 * purposes not the constant is used, but a static variable, that can be
 * changed in the test case.
 */
#define MAX_LD_ACTIVITY_SCALE (1 << 30)

/*------------------------------------------------------------------------*/
/* Here you can control, whether clauses are always or never connected.  You
 * should probably use the defaults and not uncomment any of the two lines.
 *
#define NEVER_CONNECT_LEARNED_CLAUSES
#define ALWAYS_CONNECT_LEARNED_CLAUSES
 */

/*------------------------------------------------------------------------*/

#define MAX_LEARNED_CLAUSE_SIZE UINT_MAX 		/* keep it >= 2 */
#define MAX_CONNECTED_CLAUSE_SIZE 1000
#define MAX_RELEVANCE 20
#define MAX_FORCED_CONNECTED_CLAUSE_SIZE 4
#define MAX_ALWAYS_KEPT_CLAUSE_SIZE 3

/*------------------------------------------------------------------------*/

#define MAX_CLAUSE_ACTIVITY ((1 << CEIL_LD_MAX_CLAUSE_ACTIVITY) - 1)
#define SCALE_DOWN_CLAUSE_ACTIVIES_INTERVAL 1000

/*------------------------------------------------------------------------*/
/* Maximal number of calls to 'find_unates' in 'preprocess_unates'.
 */
#define MAX_FIND_UNATES 40

/*------------------------------------------------------------------------*/
#ifdef LOG
#define REPORT_INTERVAL 1000
#define NEW_REPORT_INTERVAL(old) ((21 * (old)) / 20)
#endif
/*------------------------------------------------------------------------*/

#define GC_INTERVAL 3000
#define DUMP_FACTOR 0.20

/*------------------------------------------------------------------------*/

#define NUM_STICKY_RECENTLY_LEARNED 300

/*------------------------------------------------------------------------*/

#define RESTART_INTERVAL 10000

/*------------------------------------------------------------------------*/

#define SCALE_DOWN_VARIABLE_ACTIVITIES_INTERVAL 100

/*------------------------------------------------------------------------*/

#define BIT_MASK 7

/*------------------------------------------------------------------------*/

#define IMPLICATION_BIT 1
#define premisse2reason(lit) ((Reason*)(IMPLICATION_BIT | (Word)(lit)))
#define TOPLEVEL_REASON premisse2reason(TRUE)
#define DECISION_REASON premisse2reason(UNASSIGNED)
#define FORCED_REASON premisse2reason(FALSE)

/*------------------------------------------------------------------------*/

#define NOT_LIT(l) ((Literal*)(-(Word)(l)))
#define LITERAL(i) ((Literal*)((i) + (Variable*) 0))

/*------------------------------------------------------------------------*/

#define UNASSIGNED LITERAL(0)
#define TRUE LITERAL(1)
#define FALSE NOT_LIT(TRUE)

/*------------------------------------------------------------------------*/

#ifdef STATS
#define INCSTAT(stat,inc) do { (stat) += (inc); } while(0)
#else
#define INCSTAT(stat,inc) do { } while(0)
#endif
/*------------------------------------------------------------------------*/

#define declare_stack(scope,type,name) \
scope type* name; \
scope type* top_of_ ## name; \
scope type* end_of_ ## name

/*------------------------------------------------------------------------*/

typedef struct Chunk Chunk;
typedef struct Clause Clause;
typedef struct Frame Frame;
typedef struct GCCandidate GCCandidate;
typedef struct Link Link;
typedef struct Literal Literal;
typedef struct Binary Binary;
typedef struct Reason Reason;
typedef struct Variable Variable;
typedef struct Activity Activity;
typedef struct ActivityLink ActivityLink;
typedef struct ActivityAnchor ActivityAnchor;
typedef struct Component Component;

/*------------------------------------------------------------------------*/

enum State
{
  UNINITIALIZED_STATE,
  INITIALIZED_STATE,
  DETERMINED_SATISFIABILITY_STATE,
  DETERMINED_UNSATISFIABILITY_STATE,
  TIMED_OUT_STATE,
  DECISION_OUT_STATE,
  SPACE_OUT_STATE,
};

typedef enum State State;

/*------------------------------------------------------------------------*/

struct Chunk
{
  Chunk *next;
  size_t size;
  char data[1];			/* actually of size 'size' */
};

#define CHUNK_DATA_OFFSET ((size_t)((Chunk*)0)->data)

/*------------------------------------------------------------------------*/

struct ActivityLink
{
  Variable *prev;
  Variable *next;
};

/*------------------------------------------------------------------------*/

struct ActivityAnchor
{
  Variable *first;
  Variable *last;
};

/*------------------------------------------------------------------------*/
/* TODO: Probably we should slice variables into 3 parts.
 */
struct Variable
{
  Reason *reason;
  Literal *assignment;
  unsigned level;
#ifdef SIG
  unsigned sig;
#endif

  unsigned order;
  Component *component;

  unsigned dependent:1;
  unsigned zombie:1;
  unsigned unit:1;
  unsigned unate:1;
  unsigned sign_at_previous_visit:1;
  unsigned visited:1;

  unsigned last_active_phase:1;

  unsigned activity;
  unsigned ld_activity;
  unsigned ld_activity_scale;
  ActivityLink activity_chain;

  unsigned num_occurences[2];

  Link *clauses[2];
  Link *implications[2];
};

/*------------------------------------------------------------------------*/

struct Component
{
  unsigned start;	/* points to first variable in 'order2var' */
  unsigned end;		/* points after the last variable in 'order2var */ 
};

/*------------------------------------------------------------------------*/

struct Binary
{
  Literal *first;
  Literal *second;
};

/*------------------------------------------------------------------------*/

struct Clause
{
#if defined(TRACE) || !defined(NDEBUG)
  unsigned idx;
#endif
  Word sum;			/* sum of non false literals */
  Literal **literals;
  unsigned count;		/* counter of non false literals */
#ifdef SIG
  unsigned sig;
#endif

  unsigned connected:1;
  unsigned external:1;
  unsigned garbage:1;
  unsigned keep:1;
  unsigned used_once:1;
  unsigned used_twice:1;

  unsigned iteration;

  /* Adjust this value, if you add more bits above.
   */
#define CEIL_LD_MAX_CLAUSE_ACTIVITY 26

  unsigned activity:CEIL_LD_MAX_CLAUSE_ACTIVITY;
};

/*------------------------------------------------------------------------*/

struct GCCandidate
{
  Clause *clause;
  unsigned age;
};

/*------------------------------------------------------------------------*/

struct Link
{
  Link *tail;			/* keep this as first field */
  void *head;
};

/*------------------------------------------------------------------------*/

struct Frame
{
  unsigned satisfied;
  unsigned top_of_satisfied;
  Literal **old_top_of_trail;
};

/*------------------------------------------------------------------------*/
/******************** START OF STATIC VARIABLES PART **********************/
/*------------------------------------------------------------------------*/

static Variable variables[MAXVARS], *top_of_variables;
static unsigned max_idx;
static unsigned num_assigned;
static unsigned num_zombies, num_non_zombies;
static unsigned num_independent, num_dependent;
static int enable_cmp_var_independent;

/*------------------------------------------------------------------------*/

declare_stack(static,Variable*,order2var);
declare_stack(static,Component,components);
static unsigned active_component;

/*------------------------------------------------------------------------*/

declare_stack(static,Variable*,visited);
static unsigned num_open_paths;

/*------------------------------------------------------------------------*/

static ActivityAnchor activity_anchors[32];
static ActivityAnchor *activity_queue[32];
static unsigned max_ld_activity;
static unsigned ld_activity_scale;
static unsigned max_ld_activity_scale;
static int64 last_scale_down_variable_activities_conflicts;

/*------------------------------------------------------------------------*/

static Literal *trail[MAXVARS], **top_of_trail;
static Literal **unprocessed_clauses;
static Literal **unprocessed_implications;

/*------------------------------------------------------------------------*/

static Frame control[MAXVARS];
static Frame *top_of_control;
static unsigned decision_level;	/* == count_stack (control) */
static int64 sum_levels;
static int64 sum_length_learned_clauses;
#ifdef SIG
static int64 sum_length_learned_sigs;
static int64 sum_learned_sigs;
#endif

/*------------------------------------------------------------------------*/

static Literal *pushed_literals[MAXVARS];
static Literal **top_of_pushed_literals;

/*------------------------------------------------------------------------*/

declare_stack(static,Clause*,clauses);

static unsigned first_learned_clause_idx;
static int valid_first_learned_clause_idx;
static unsigned num_clauses;
static unsigned num_empty_clauses;
static unsigned num_binary_clauses;
static unsigned num_implications;
static unsigned num_learned_clauses;
static int64 last_scale_down_clause_activities_conflicts;
#if defined(TRACE) || !defined(NDEBUG)
static unsigned next_clause_idx;
#endif

/*------------------------------------------------------------------------*/

declare_stack(static,Literal*,units);
declare_stack(static,Literal*,unates);
static unsigned old_count_units, old_count_unates;

/*------------------------------------------------------------------------*/

static FILE *out;
static int close_out;
static FILE *in;
static const char *name_in;
static int close_in;

/*------------------------------------------------------------------------*/

static double seconds;

/*------------------------------------------------------------------------*/

static int64 decision_limit;
static int space_limit;		/* in MB */
static double time_limit;	/* in seconds */

/*------------------------------------------------------------------------*/

static int64 conflicts;
static int64 decisions;
static unsigned iterations;

/*------------------------------------------------------------------------*/

static int64 sum_learned_clauses;
static int64 sum_connected_clauses;
static int64 sum_multiple_used_learned_clauses;
static int64 sum_multiple_used_connected_clauses;

/*------------------------------------------------------------------------*/

static int64 restarts;
static unsigned restart_interval;
static int64 last_restart_conflicts;

/*------------------------------------------------------------------------*/

static int64 gcs;
static int64 last_gc_conflicts;
static double dump_factor;

/*------------------------------------------------------------------------*/
#ifdef TRACE
static char reason2str_buffer[40];
#endif
/*------------------------------------------------------------------------*/
#ifdef LOG
static int64 reports;
static unsigned report_interval;
static int64 last_report_decisions;
#endif
/*------------------------------------------------------------------------*/
#ifdef STATS
/*------------------------------------------------------------------------*/

static int64 adjustments;
static int64 adjusted_levels;
static int64 assignments;
static int64 backtracks;
static int64 collected;
static int64 decisions_based_on_activity;
static int64 hits;
static int64 implications;
static int64 irrelevant;
static int64 lookups;
static int64 skipped_backtracks;
static int64 top_level_satisfied_clauses;
static int64 top_level_assignments;
static int64 uips;
static int64 visits;

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

static Chunk *chunks;
static char *mem, *top_of_mem, *end_of_mem;
static size_t allocated_mem;
static size_t max_allocated_mem;
static size_t dynamic_allocated_bytes;
static size_t max_dynamic_allocated_bytes;
static size_t static_allocated_bytes;

/*------------------------------------------------------------------------*/

static double entered_time;

/*------------------------------------------------------------------------*/

static unsigned entered = 0;
static State state = UNINITIALIZED_STATE;

/*------------------------------------------------------------------------*/
/********************** END OF STATIC VARIABLES PART **********************/
/*------------------------------------------------------------------------*/

#define count_stack(stack) \
(top_of_ ## stack - stack)

/*------------------------------------------------------------------------*/

#define size_stack(stack) \
(end_of_ ## stack - stack)

/*------------------------------------------------------------------------*/

#define init_stack(stack) \
do { end_of_ ## stack = top_of_ ## stack = stack = 0; } while(0)

/*------------------------------------------------------------------------*/

#define release_stack(stack) \
do { \
  size_t bytes = size_stack(stack); \
  bytes *= sizeof (stack[0]); \
  assert (dynamic_allocated_bytes >= bytes); \
  nanosat_free (stack, bytes); \
  init_stack (stack); \
} \
while(0)

/*------------------------------------------------------------------------*/
/* Certain stacks can be resized dynamically.  Their memory management is on
 * the same level as the low level allocator 'nanosat_{malloc,free}'.
 */
static void
enlarge_stack (size_t elem_size,
	       void *start_ptr, void *top_ptr, void *end_ptr)
{
  size_t old_size, new_size, *start, *old_start;
  size_t *real_old_start, real_new_size;
  char *top, *end;

  start = *(void **) start_ptr;
  top = *(void **) top_ptr;
  end = *(void **) end_ptr;
  real_old_start = old_start = start;
  old_size = ((char*)end) - (char*)start;
  real_new_size = new_size = old_size ? 2 * old_size : elem_size;
#ifndef NDEBUG
  real_new_size += sizeof(size_t);
  if (real_old_start)
    {
      real_old_start--;
      assert (*real_old_start == old_size);
    }
  else
    assert (!old_size);
#endif
  start = realloc (real_old_start, real_new_size);
#ifndef NDEBUG
  *start++ = new_size;
#endif
  end = ((char*)start) + new_size;
  top = top + (((char*)start) - (char*)old_start);

  *(void **) start_ptr = start;
  *(void **) top_ptr = top;
  *(void **) end_ptr = end;

  assert (new_size > old_size);

  dynamic_allocated_bytes += new_size - old_size;
  if (dynamic_allocated_bytes > max_dynamic_allocated_bytes)
    max_dynamic_allocated_bytes = dynamic_allocated_bytes;
}

/*------------------------------------------------------------------------*/

#define push_stack(stack,elem) \
do { \
  if (top_of_ ## stack >= end_of_ ## stack) \
    enlarge_stack (sizeof(stack[0]), \
	           &stack, &top_of_ ## stack, &end_of_ ## stack); \
  *top_of_ ## stack++ = (elem); \
} \
while(0)

/*------------------------------------------------------------------------*/
/* This is the allocation function of the low level memory manage, which
 * is just a wrapper around 'malloc/free' adding resource usage monitoring.
 */
static void *
nanosat_malloc (size_t bytes)
{
  size_t *res, real_bytes = bytes;

  dynamic_allocated_bytes += bytes;
#ifndef NDEBUG
  real_bytes += sizeof(*res);
#endif
  res = malloc (real_bytes);
#ifndef NDEBUG
  *res++ = bytes;
#endif

  return res;
}

/*------------------------------------------------------------------------*/
/* This is the deallocation function of the low level memory allocator.
 */
static void
nanosat_free (void *ptr, size_t bytes)
{
  size_t * real_ptr = ptr;

#ifndef NDEBUG
  assert (!ptr == !bytes);
  if (real_ptr)
    {
      real_ptr--;
      assert (*real_ptr == bytes);
    }
#endif
  assert (dynamic_allocated_bytes >= bytes);
  dynamic_allocated_bytes -= bytes;
  free (real_ptr);
}

/*------------------------------------------------------------------------*/
/* The medium layer of the memory allocation stack operates with chunks of
 * minimum size 'MIN_CHUNK_SIZE'.  The actual size of a new chunk of memory
 * will be at least 'min_size'.  This allows the moving garbage collector to
 * allocate one contigous block for the target memory and also cares for
 * very large clauses.  After a chunk is allocated it is set up to provide
 * space for the simple stack allocator 'new'.
 */
static void
new_chunk (size_t min_size)
{
  size_t bytes, size;
  Chunk *chunk;

  assert (top_of_mem + min_size > end_of_mem);

  size = (min_size <= MIN_CHUNK_SIZE) ? MIN_CHUNK_SIZE : min_size;
  bytes = size + CHUNK_DATA_OFFSET;

  chunk = nanosat_malloc (bytes);
  chunk->size = size;

  chunk->next = chunks;
  chunks = chunk;

  mem = chunk->data;
  top_of_mem = mem;
  end_of_mem = mem + size;

  memset (mem, 0, size);

  allocated_mem += size;
  if (allocated_mem > max_allocated_mem)
    max_allocated_mem = allocated_mem;
}

/*------------------------------------------------------------------------*/

static void
delete_chunk (Chunk * chunk)
{
  assert (allocated_mem >= chunk->size);
  allocated_mem -= chunk->size;
  nanosat_free (chunk, chunk->size + CHUNK_DATA_OFFSET);
}

/*------------------------------------------------------------------------*/
/* Chunks are linked chronologically through their next field:  the chunk
 * pointed to by 'next' has been allocated earlier.  With 'delete_chunks' we
 * delete all chunks older and including the 'chunk' argument.
 */
static void
delete_chunks (Chunk * chunk)
{
  Chunk *p, *next;

  for (p = chunk; p; p = next)
    {
      next = p->next;
      delete_chunk (p);
    }
}

/*------------------------------------------------------------------------*/
/* Memory allocated with new is supposed to be under control of the moving
 * garbage collector.   In particular it is deleted after the 'move' phase of
 * the garbage collector.  Currently we only allocate parts of the clause
 * database with 'new', which consists of links, literals, and clauses.
 */
static void *
new (size_t bytes)
{
  char *new_top_of_mem;
  void *res;

  assert (bytes > 0);

  new_top_of_mem = top_of_mem + bytes;
  if (new_top_of_mem > end_of_mem)
    {
      new_chunk (bytes);
      new_top_of_mem = mem + bytes;
      assert (new_top_of_mem <= end_of_mem);
    }

  res = top_of_mem;
  top_of_mem = new_top_of_mem;

  return res;
}

/*------------------------------------------------------------------------*/

static Link *
cons (void *head, Link * tail)
{
  Link *res;

  res = (Link *) new (sizeof (*res));
  res->tail = tail;
  res->head = head;

  return res;
}

/*------------------------------------------------------------------------*/

static Link *
cdr (Link * link)
{
  assert (link);
  return link->tail;
}

/*------------------------------------------------------------------------*/

static void *
car (Link * link)
{
  assert (link);
  return link->head;
}

/*------------------------------------------------------------------------*/

static int
lit2sign (Literal * lit)
{
  return (0 < (Word) lit) ? 1 : -1;
}

/*------------------------------------------------------------------------*/

static Literal *
not_lit (Literal * lit)
{
  return NOT_LIT (lit);
}

/*------------------------------------------------------------------------*/

static Literal *
var2lit (Variable * v, int sign)
{
  Literal *res;
  Word tmp;

  assert (sign == -1 || sign == 1);

  tmp = (Word) v;
  tmp *= sign;
  res = (Literal *) tmp;

  return res;
}

/*------------------------------------------------------------------------*/

static Variable *
lit2var (Literal * lit)
{
  Variable *res;
  Word tmp;

  tmp = (Word) lit;
  if (tmp < 0)
    tmp = -tmp;

  res = (Variable *) tmp;

  return res;

}

/*------------------------------------------------------------------------*/

static unsigned
sign2idx (int sign)
{
  assert (sign == -1 || sign == 1);
  return (sign < 0);
}

/*------------------------------------------------------------------------*/

static int
idx2sign (unsigned idx)
{
  assert (idx == 0 || idx == 1);
  return idx ? -1 : 1;
}

/*------------------------------------------------------------------------*/

static Link **
lit2implicationsptr (Literal * lit)
{
  Link **res;
  Variable *v;
  int sign;

  v = lit2var (lit);
  sign = lit2sign (lit);
  res = v->implications;
  res += sign2idx (sign);

  return res;
}

/*------------------------------------------------------------------------*/

static Link **
lit2clausesptr (Literal * lit)
{
  Variable *v;
  Link **res;
  int sign;

  v = lit2var (lit);
  sign = lit2sign (lit);
  res = v->clauses;
  res += sign2idx (sign);

  return res;
}

/*------------------------------------------------------------------------*/

static void
push_lit (Literal * lit)
{
  *top_of_pushed_literals++ = lit;
}

/*------------------------------------------------------------------------*/

static int
int2sign (int a)
{
  return (a < 0) ? -1 : 1;
}

/*------------------------------------------------------------------------*/

static int
int2nat (int a)
{
  return (a < 0) ? -a : a;
}

/*------------------------------------------------------------------------*/

static void
merge_activity_chains (ActivityAnchor * dst, ActivityAnchor * src)
{
  if (!src->first)
    return;

  assert (src->last);
  assert (!src->last->activity_chain.next);

  src->last->activity_chain.next = dst->first;
  if (dst->first)
    {
      assert (!dst->first->activity_chain.prev);
      dst->first->activity_chain.prev = src->last;
    }
  else
    {
      assert (!dst->last);
      dst->last = src->last;
    }

  dst->first = src->first;
}

/*------------------------------------------------------------------------*/

static void
print (FILE * file, const char *fmt, ...)
{
  va_list ap;

  fputs (NANOSAT_PREFIX, out);
  va_start (ap, fmt);
  vfprintf (out, fmt, ap);
  va_end (ap);
  fputc ('\n', out);
  fflush (out);
}

/*------------------------------------------------------------------------*/
#ifdef LOG
/*------------------------------------------------------------------------*/

#define log1(a) print (out,a)
#define log2(a,b) print (out,a,b)
#define log3(a,b,c) print (out,a,b,c)
#define log4(a,b,c,d) print (out,a,b,c,d)
#define log5(a,b,c,d,e) print (out,a,b,c,d,e)
#define log6(a,b,c,d,e,f) print (out,a,b,c,d,e,f)
#define log7(a,b,c,d,e,f,g) print (out,a,b,c,d,e,f,g)
#define log8(a,b,c,d,e,f,g,h) print (out,a,b,c,d,e,f,g,h)
#define log9(a,b,c,d,e,f,g,h,i) print (out,a,b,c,d,e,f,g,h,i)

/*------------------------------------------------------------------------*/
#else
/*------------------------------------------------------------------------*/

#define log1(a) do { } while(0)
#define log2(a,b) do { } while(0)
#define log3(a,b,c) do { } while(0)
#define log4(a,b,c,d) do { } while(0)
#define log5(a,b,c,d,e) do { } while(0)
#define log6(a,b,c,d,e,f) do { } while(0)
#define log7(a,b,c,d,e,f,g) do { } while(0)
#define log8(a,b,c,d,e,f,g,h) do { } while(0)
#define log9(a,b,c,d,e,f,g,h,i) do { } while(0)

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/
#ifdef TRACE
/*------------------------------------------------------------------------*/

#define trace1(a) print (out,"%u " a,decision_level)
#define trace2(a,b) print (out,"%u " a,decision_level,b)
#define trace3(a,b,c) print (out,"%u " a,decision_level,b,c)
#define trace4(a,b,c,d) print (out,"%u " a,decision_level,b,c,d)
#define trace5(a,b,c,d,e) print (out,"%u " a,decision_level,b,c,d,e)

/*------------------------------------------------------------------------*/
#else
/*------------------------------------------------------------------------*/

#define trace1(a) do { } while(0)
#define trace2(a,b) do { } while(0)
#define trace3(a,b,c) do { } while(0)
#define trace4(a,b,c,d) do { } while(0)
#define trace5(a,b,c,d,e) do { } while(0)

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/
#if !defined(NDEBUG) || defined(TEST)
/*------------------------------------------------------------------------*/

static unsigned
ldfloor (unsigned n)
{
  int i;

  for (i = 0; i < 31; i++)
    if (n <= (1 << (i + 1)) - 1)
      break;

  return i;
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

static unsigned
max_reverse_ldfloor (unsigned n)
{
  assert (n < 32);

  return (1 << (n + 1)) - 1;
}

/*------------------------------------------------------------------------*/

static unsigned
ldceil (unsigned n)
{
  int i;

  for (i = 0; i <= 31; i++)
    if (n <= (1 << i))
      break;

  return i;
}

/*------------------------------------------------------------------------*/

static void
normalize_variable_activity (Variable * v)
{
  unsigned activity, ld_activity;
  int s;

  assert (v->ld_activity_scale <= ld_activity_scale);

  if (v->ld_activity_scale == ld_activity_scale)
    return;

  s = ld_activity_scale - v->ld_activity_scale;
  assert (s > 0);

  activity = v->activity;
  activity = (s >= 32) ? 0 : (activity >> s);

  ld_activity = v->ld_activity;
  ld_activity = (ld_activity <= s) ? 0 : ld_activity - s;

  assert (ldfloor (activity) == ld_activity);
  v->activity = activity;
  v->ld_activity = ld_activity;

  v->ld_activity_scale = ld_activity_scale;
}

/*------------------------------------------------------------------------*/

static void
reset_ld_activity_scale (void)
{
  Variable *v;

  for (v = variables + 1; v < top_of_variables; v++)
    {
      normalize_variable_activity (v);
      v->ld_activity_scale = 0;
    }

  ld_activity_scale = 0;

  log2 ("reset of activity scale to ld_activity_scale=%u", ld_activity_scale);
}

/*------------------------------------------------------------------------*/

static void
scale_down_variable_activities (void)
{
  ActivityAnchor *tmp;
  ActivityAnchor **p;

  trace1 ("scaling down variable activities");
  trace3 ("  old ld_activity_scale=%u old max_ld_activity=%u",
	  ld_activity_scale, max_ld_activity);

  if (max_ld_activity > 0)
    {
      merge_activity_chains (activity_queue[0], activity_queue[1]);

      tmp = activity_queue[1];

      for (p = activity_queue + 1; p < activity_queue + max_ld_activity; p++)
	p[0] = p[1];

      activity_queue[max_ld_activity] = tmp;
      tmp->first = tmp->last = 0;
      max_ld_activity--;
    }

  ld_activity_scale++;
  if (ld_activity_scale > max_ld_activity_scale)
    reset_ld_activity_scale ();

  trace3 ("  new ld_activity_scale=%u new max_ld_activity=%u",
	  ld_activity_scale, max_ld_activity);
}

/*------------------------------------------------------------------------*/

static unsigned
normalized_ld_activity (Variable * v)
{
  unsigned res, scale;

  assert (v->ld_activity_scale <= ld_activity_scale);

  scale = ld_activity_scale - v->ld_activity_scale;
  res = v->ld_activity;
  res = (scale >= res) ? 0 : res - scale;

  assert (res < 32);

  return res;
}

/*------------------------------------------------------------------------*/

static ActivityAnchor *
var2anchor (Variable * v)
{
  ActivityAnchor * res;
  unsigned ld_activity;

  ld_activity = normalized_ld_activity (v);
  res = activity_queue[ld_activity];

  return res;
}

/*------------------------------------------------------------------------*/

static void
dequeue_from_activity_queue (Variable * v)
{
  ActivityAnchor * anchor;
  Variable *prev, *next;

  anchor = var2anchor (v);

  prev = v->activity_chain.prev;
  next = v->activity_chain.next;

  if (prev)
    {
      prev->activity_chain.next = next;
    }
  else
    {
      assert (anchor->first == v);
      anchor->first = next;
    }

  if (next)
    {
      next->activity_chain.prev = prev;
    }
  else
    {
      assert (anchor->last == v);
      anchor->last = prev;
    }
}

/*------------------------------------------------------------------------*/

static void
enqueue_on_activity_queue (Variable * v)
{
  ActivityAnchor * anchor;
  unsigned ld_activity;
  Variable *first;

  ld_activity = normalized_ld_activity (v);
  if (ld_activity > max_ld_activity)
    max_ld_activity = ld_activity;

  anchor = activity_queue [ld_activity];
  first = anchor->first;

  v->activity_chain.prev = 0;
  v->activity_chain.next = first;

  if (first)
    {
      assert (!first->activity_chain.prev);
      first->activity_chain.prev = v;
    }
  else
    {
      assert (!anchor->last);
      anchor->last = v;
    }

  anchor->first = v;
}

/*------------------------------------------------------------------------*/

static void
inc_variable_activity (Variable * v)
{
  unsigned ld_new_activity;

  normalize_variable_activity (v);

  v->activity = v->activity + 1;
  v->ld_activity_scale = ld_activity_scale;

  if (v->activity <= max_reverse_ldfloor (v->ld_activity))
    return;

  if (v->assignment == UNASSIGNED)
    dequeue_from_activity_queue (v);

  ld_new_activity = v->ld_activity + 1;
  v->ld_activity = ld_new_activity;
  assert (v->ld_activity == ldfloor (v->activity));

  if (v->assignment == UNASSIGNED)
    enqueue_on_activity_queue (v);
}

/*------------------------------------------------------------------------*/

static void
scale_down_clause_activities (void)
{
  unsigned old_activity, new_activity;
  Clause **p, * clause;

  trace1 ("scaling down clause activities");

  for (p = clauses; p < top_of_clauses; p++)
    {
      clause = *p;
      old_activity = clause->activity;
      new_activity = old_activity >> 1;
      clause->activity = new_activity;
    }
}

/*------------------------------------------------------------------------*/

static void
inc_clause_activity (Clause * clause)
{
  unsigned old_activity, new_activity;

  if (clause->activity >= MAX_CLAUSE_ACTIVITY)
    scale_down_clause_activities ();

  old_activity = clause->activity;
  new_activity = old_activity + 1;
  clause->activity = new_activity;
}

/*------------------------------------------------------------------------*/

static void
inc_clause_usage (Clause * clause)
{
  if (!clause->used_once)
    clause->used_once = 1;
  else if (!clause->used_twice)
    {
      clause->used_twice = 1;

      if (!clause->external)
	{
	  sum_multiple_used_learned_clauses++;
   
	  if (clause->connected)
	    sum_multiple_used_connected_clauses++;
	}
    }
}

/*------------------------------------------------------------------------*/

static void
init_vars (Variable * start, Variable * end)
{
  Variable *p;

  memset (start, 0, ((char *) end) - (char *) start);

  if (start == variables)
    start++;

  for (p = start; p < end; p++)
    enqueue_on_activity_queue (p);
}

/*------------------------------------------------------------------------*/

static int
var2idx (Variable * v)
{
  assert (v >= variables);
  assert (v < top_of_variables);

  return v - variables;
}

/*------------------------------------------------------------------------*/

static Variable *
idx2var (int idx)
{
  assert (idx > 0);
  assert (idx < count_stack (variables));

  return variables + idx;
}

/*------------------------------------------------------------------------*/

static void
push_int_lit (int int_lit)
{
  Variable *new_eov, *v;
  Literal *lit;
  unsigned idx;
  int sign;

  sign = int2sign (int_lit);
  idx = int2nat (int_lit);
  assert (idx > 0);
  v = variables + idx;

  if (v >= top_of_variables)
    {
      assert (idx >= top_of_variables - variables);
      new_eov = variables + idx + 1;
      init_vars (top_of_variables, new_eov);
      top_of_variables = new_eov;
      max_idx = idx;
    }

  lit = var2lit (v, sign);
  push_lit (lit);
}

/*------------------------------------------------------------------------*/

static Literal *
int2lit (int lit)
{
  Literal * res;
  unsigned idx;
  Variable * v;
  int sign;

  sign = int2sign (lit);
  idx = int2nat (lit);
  v = idx2var (idx);
  res = var2lit (v, sign);

  return res;
}

/*------------------------------------------------------------------------*/

static Word *
literals2header (Literal ** literals)
{
  return (Word *) literals - 1;
}

/*------------------------------------------------------------------------*/

static Literal **
header2literals (Word * header)
{
  return (Literal **) header + 1;
}

/*------------------------------------------------------------------------*/

static unsigned
size_literals (Literal ** literals)
{
  Word *header;
  unsigned res;

  header = literals2header (literals);
  res = *header;

  return res;
}

/*------------------------------------------------------------------------*/

static size_t
bytes_literals (unsigned size)
{
  return sizeof (Literal *) * (1 + size);
}

/*------------------------------------------------------------------------*/

static Literal **
new_literals (unsigned size)
{
  Literal **res;
  Word *header;
  size_t bytes;

  bytes = bytes_literals (size);
  header = (Word *) new (bytes);
  *header = size;
  res = header2literals (header);

  return res;
}

/*------------------------------------------------------------------------*/

static Literal **
realloc_literals (Literal ** literals)
{
  Word *old_header, *new_header;
  Literal **res;
  unsigned size;
  size_t bytes;

  size = size_literals (literals);
  bytes = bytes_literals (size);
  old_header = literals2header (literals);
  new_header = (Word *) new (bytes);
  memcpy (new_header, old_header, bytes);
  res = (Literal **) (new_header + 1);

  return res;
}

/*------------------------------------------------------------------------*/

static void
push_empty_clause (void)
{
  num_empty_clauses++;
}

/*------------------------------------------------------------------------*/

static int
lit2idx (Literal * lit)
{
  Variable *v;
  int res;

  v = lit2var (lit);
  res = var2idx (v);

  return res;
}

/*------------------------------------------------------------------------*/

static int
lit2int (Literal * lit)
{
  int idx, sign, res;

  idx = lit2idx (lit);
  sign = lit2sign (lit);
  res = sign * idx;

  return res;
}

/*------------------------------------------------------------------------*/

static void
inc_num_occurences (Literal * lit)
{
  unsigned sign_idx = sign2idx (lit2sign (lit));
  Variable *v = lit2var (lit);

  v->num_occurences[sign_idx] += 1;
  assert (v->num_occurences[sign_idx]);
}

/*------------------------------------------------------------------------*/

static void
inc_num_occurences_of_pushed_literals (void)
{
  Literal **p;

  for (p = pushed_literals; p < top_of_pushed_literals; p++)
    inc_num_occurences (*p);
}

/*------------------------------------------------------------------------*/

static void
push_unit_clause (void)
{
  Literal *lit;
  Variable * v;

  assert (count_stack (pushed_literals) == 1);

  lit = *pushed_literals;
  v = lit2var (lit);
  push_stack(units, lit);
  v->unit = 1;

  trace2 ("unit(%d)", lit2int (lit));
}

/*------------------------------------------------------------------------*/

static void
push_implication (Literal * a, Literal * b)
{
  Link **implications;

  implications = lit2implicationsptr (a);
  *implications = cons (b, *implications);
  num_implications++;

  trace3 ("implication(%d,%d)", lit2int (not_lit (a)), lit2int (b));
}

/*------------------------------------------------------------------------*/

static void
push_binary_clause (void)
{
  Literal *first, *second;

  assert (count_stack (pushed_literals) == 2);

  first = pushed_literals[0];
  second = pushed_literals[1];

  push_implication (first, second);
  push_implication (second, first);

  num_binary_clauses++;
}

/*------------------------------------------------------------------------*/

static Literal *
deref (Literal * lit)
{
  Literal *res;
  Variable *v;
  int sign;

  v = lit2var (lit);
  res = v->assignment;
  sign = lit2sign (lit);
  if (sign < 0)
    res = not_lit (res);

  return res;
}

/*------------------------------------------------------------------------*/

static void
add_literals (Clause * clause, Literal ** literals, unsigned size)
{
  Literal *lit, **p, **q, **top_of_literals;

  assert (size > 2);
  p = literals;
  top_of_literals = literals + size;
  q = clause->literals = new_literals (size);

  while (p < top_of_literals)
    {
      lit = *p++;
      *q++ = lit;
    }
  assert (q == clause->literals + size);
}

/*------------------------------------------------------------------------*/

static void
connect_literals (Clause * clause)
{
  Literal *lit, **p, **top_of_literals, *tmp;
  Link **clausesptr;

  clause->count = 0;
  clause->sum = 0;

  p = clause->literals;
  top_of_literals = p + size_literals (p);
  assert (size_literals (p) > 2);

  while (p < top_of_literals)
    {
      lit = *p++;

      clausesptr = lit2clausesptr (lit);
      *clausesptr = cons (clause, *clausesptr);

      tmp = deref (lit);
      if (tmp == FALSE)
	continue;

      clause->count++;
      clause->sum ^= (Word) lit;
    }

  clause->connected = 1;

  trace2 ("connect(%u)", clause->idx);
}

/*------------------------------------------------------------------------*/

static void
trace_clause (Clause * clause)
{
#ifdef TRACE
  Literal **p, **top_of_literals;

  fprintf (out, "%s%u clause(%u) =", 
           NANOSAT_PREFIX, decision_level, clause->idx);

  p = clause->literals;
  top_of_literals = p + size_literals (p);
  while (p < top_of_literals)
    fprintf (out, " %d", lit2int (*p++));

  fputc ('\n', out);
#endif
}

/*------------------------------------------------------------------------*/

static Clause *
push_large_clause (int connect)
{
  unsigned size;
  Clause *res;

  res = (Clause *) new (sizeof (*res));
  memset (res, 0, sizeof (*res));
#if defined(TRACE) || !defined(NDEBUG)
  res->idx = next_clause_idx++;
  assert (next_clause_idx);
#endif
  res->iteration = iterations;
  size = count_stack (pushed_literals);
  add_literals (res, pushed_literals, size);
  if (connect)
    connect_literals (res);
  push_stack (clauses, res);
  num_clauses++;

  trace_clause (res);

  return res;
}

/*------------------------------------------------------------------------*/

static Clause *
push_clause (int connect)
{
  Clause * res;
  Word size;

  size = top_of_pushed_literals - pushed_literals;

  inc_num_occurences_of_pushed_literals ();
  res = 0;

  if (!size)
    push_empty_clause ();
  else if (size == 1)
    push_unit_clause ();
  else if (size == 2)
    push_binary_clause ();
  else
    res = push_large_clause (connect);

  top_of_pushed_literals = pushed_literals;

  return res;
}

/*------------------------------------------------------------------------*/
#ifndef NDEBUG
/*------------------------------------------------------------------------*/

static int
is_simplified_clause (void)
{
  Literal **p;
  Variable *v;
  int res;

  res = 1;

  for (p = pushed_literals; res && p < top_of_pushed_literals; p++)
    {
      v = lit2var (*p);
      res = !v->visited;
      v->visited = 1;
    }

  while (p > pushed_literals)
    lit2var (*--p)->visited = 0;

  return res;
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

static int
is_trivial_clause (void)
{
  int sign, sign_idx, res;
  Literal **p, *lit;
  Variable *v;

  res = 0;

  for (p = pushed_literals; !res && p < top_of_pushed_literals; p++)
    {
      lit = *p;
      v = lit2var (lit);
      sign = lit2sign (lit);
      sign_idx = sign2idx (sign);

      if (v->visited)
	{
	  if (sign_idx != v->sign_at_previous_visit)
	    res = 1;
	}
      else
	{
	  v->visited = 1;
	  if (sign_idx)
	    v->sign_at_previous_visit = 1;
	}
    }

  while (p > pushed_literals)
    {
      v = lit2var (*--p);
      v->visited = 0;
      v->sign_at_previous_visit = 0;
    }

  return res;
}

/*------------------------------------------------------------------------*/

static void
simplify_clause (void)
{
  Literal **p, **q, *lit;
  Variable * v;

  q = pushed_literals;

  for (p = pushed_literals; p < top_of_pushed_literals; p++)
    {
      lit = *p;
      v = lit2var (lit);
      if (v->visited)
	continue;

      v->visited = 1;
      *q++ = lit;
    }

  top_of_pushed_literals = q;

  while (q-- > pushed_literals)
    lit2var(*q)->visited = 0;
}

/*------------------------------------------------------------------------*/

static void
push_external_clause (void)
{
  Clause * clause;

  if (!is_trivial_clause ())
    {
      simplify_clause ();
      clause = push_clause (1);
      if (clause)
	clause->external = 1;
    }
  else
    top_of_pushed_literals = pushed_literals;
}

/*------------------------------------------------------------------------*/

static unsigned
avg_len ()
{
  unsigned res;
  int64 tmp;

  if (!sum_learned_clauses)
    return 0;

  tmp = sum_length_learned_clauses + sum_learned_clauses - 1;
  tmp /= sum_learned_clauses;

  assert (tmp < (1u << 31));
  res = tmp;

  return res;
}

/*------------------------------------------------------------------------*/
#if defined(LOG) || defined(STATS)
/*------------------------------------------------------------------------*/

static double
avg_len_as_double ()
{
  if (!sum_learned_clauses)
    return 0;

  return sum_length_learned_clauses / (double) sum_learned_clauses;
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

static unsigned
max_connected_clause_size (void)
{
  unsigned res;
#ifdef ALWAYS_CONNECT_LEARNED_CLAUSES
  return UINT_MAX;
#elif defined(NEVER_CONNECT_LEARNED_CLAUSES)
  return 0;
#else
  res = avg_len ();

  if (res < MAX_FORCED_CONNECTED_CLAUSE_SIZE)
    res = MAX_FORCED_CONNECTED_CLAUSE_SIZE;
  else if (res > MAX_CONNECTED_CLAUSE_SIZE)
    res = MAX_CONNECTED_CLAUSE_SIZE;

  return res;
#endif
}

/*------------------------------------------------------------------------*/

static void
push_learned_clause (void)
{
  unsigned size, connect;
  
  assert (is_simplified_clause());

  size = top_of_pushed_literals - pushed_literals;

  if (size <= 2)
    {
      push_clause (1);
    }
  else if (size <= MAX_LEARNED_CLAUSE_SIZE)
    {
      if (!valid_first_learned_clause_idx)
	{
	  valid_first_learned_clause_idx = 1;
	  first_learned_clause_idx = count_stack (clauses);
	}

      connect = (size <= max_connected_clause_size());
      push_clause (connect);

      if (connect)
	sum_connected_clauses++;

      num_learned_clauses++;

      sum_learned_clauses++;
      sum_length_learned_clauses += size;
    }
  else
    top_of_pushed_literals = pushed_literals;
}

/*------------------------------------------------------------------------*/

static int
parse (void)
{
  int ch, count, sign, idx, lineno, v, c;
  unsigned parsed_clauses;

  sign = 1;
  parsed_clauses = 0;
  lineno = 1;
  count = 0;
  v = -1;
  c = -1;

PARSE0:

  ch = getc (in);

PARSE1:

  if (ch == '\n')
    lineno++;

  if (ch == EOF)
    {
      if (count)
	{
	  fprintf (out, "%s.%d: '0' missing\n", name_in, lineno);
	  return 0;
	}

      if (v < 0)
	{
	  fprintf (out, "%s.%d: 'p cnf ...' not found\n", name_in, lineno);
	  return 0;
	}

      if (v < max_idx)
	{
	  fprintf (out, "%s.%d: specified index exceeded\n", name_in, lineno);
	  return 0;
	}

      if (c != parsed_clauses)
	{
	  fprintf (out, "%s.%d: mismatched specified clauses\n", name_in,
		   lineno);
	  return 0;
	}

      return 1;
    }

  if (ch == ' ' || ch == '\t' || ch == '\n')
    goto PARSE0;

  if (ch == 'c')
    {
      while ((ch = getc (in)) != '\n' && ch != EOF)
	;

      goto PARSE1;
    }

  if (ch == 'p')
    {
      int tmp;
      if ((tmp = fscanf (in, " cnf %u %u", &v, &c)) != 2)
	{
	  fprintf (out, "%s.%d: invalid 'p ...' specification\n", name_in,
		   lineno);
	  return 0;
	}

      goto PARSE0;
    }

  if (ch == '-')
    {
      sign = -1;
      ch = fgetc (in);
    }

  if (!isdigit (ch))
    {
      fprintf (out, "%s.%d: unexpected character '%c'\n", name_in, lineno,
	       ch);
      return 0;
    }

  idx = ch - '0';

PARSE2:

  ch = getc (in);
  if (isdigit (ch))
    {
      idx = 10 * idx + (ch - '0');
      goto PARSE2;
    }

  if (idx >= MAXVARS)
    {
      fprintf (out, "%s.%d: maximal number of variables exceeded\n",
	       name_in, lineno);
      return 0;
    }

  if (idx)
    {
      push_int_lit (sign * idx);
      count++;
    }
  else
    {
      push_external_clause ();
      parsed_clauses++;
      count = 0;
    }

  sign = 1;

  goto PARSE1;
}

/*------------------------------------------------------------------------*/

static void
dump_clause (Clause * clause, FILE * file)
{
  Literal **p, **eor, *lit;

  file = !file ? stdout : file;

  p = clause->literals;
  eor = p + size_literals (p);

  while (p < eor)
    {
      lit = *p++;
      assert (lit);
      fprintf (file, "%d ", lit2int (lit));
    }

  fprintf (file, "0\n");
}


/*------------------------------------------------------------------------*/

static void
dump_clauses (FILE * file)
{
  Clause **p;

  for (p = clauses; p < top_of_clauses; p++)
    dump_clause (*p, file);
}

/*------------------------------------------------------------------------*/

static void
dump_empty_clauses (FILE * file)
{
  unsigned i;

  for (i = 0; i < num_empty_clauses; i++)
    fprintf (file, "0\n");
}

/*------------------------------------------------------------------------*/

static void
dump_units (FILE * file)
{
  Literal **p;

  for (p = units; p < top_of_units; p++)
    fprintf (file, "%d 0\n", lit2int (*p));
}

/*------------------------------------------------------------------------*/

static unsigned
length_cnf (void)
{
  unsigned res;

  res = num_clauses;
  res += num_empty_clauses;
  res += count_stack (units);
  res += num_binary_clauses;

  return res;
}

/*------------------------------------------------------------------------*/

static void
dump_implications (FILE * file)
{
  Literal *this, *other;
  int this_as_int, sign;
  Variable *u, *v;
  Link *p;

  for (u = variables; u < top_of_variables; u++)
    {
      for (sign = -1; sign <= 1; sign += 2)
	{
	  this = var2lit (u, sign);
	  this_as_int = lit2int (this);

	  for (p = *lit2implicationsptr (this); p; p = cdr (p))
	    {
	      other = car (p);
	      v = lit2var (other);
	      assert (u != v);
	      if (u > v)
		continue;

	      fprintf (file, "%d %d 0\n", this_as_int, lit2int (other));
	    }
	}
    }
}

/*------------------------------------------------------------------------*/

static void
dump (FILE * file, int include_header)
{
  if (include_header)
    fprintf (file, "p cnf %.0f %.0f\n", (double) max_idx,
	     (double) length_cnf());

  dump_empty_clauses (file);
  dump_units (file);
  dump_implications (file);
  dump_clauses (file);
}

/*------------------------------------------------------------------------*/

static int
is_assigned (Literal * lit)
{
  return lit2var (lit)->assignment != UNASSIGNED;
}

/*------------------------------------------------------------------------*/

static Literal *
sign2const (int sign)
{
  Literal *res;
  Word tmp;

  assert (sign == -1 || sign == 1);

  tmp = (Word) TRUE;
  tmp *= sign;
  res = (Literal *) tmp;

  return res;
}

/*------------------------------------------------------------------------*/

static Reason *
clause2reason (Clause * clause)
{
  return (Reason *) clause;
}

/*------------------------------------------------------------------------*/

static int
is_implication (Reason * reason)
{
  return (IMPLICATION_BIT & (Word) reason) != 0;
}

/*------------------------------------------------------------------------*/

static int
is_two_valued_constant (Literal * lit)
{
  return lit == TRUE || lit == FALSE;
}

/*------------------------------------------------------------------------*/
#if 0

static int
is_three_valued_constant (Literal * lit)
{
  return lit == TRUE || lit == FALSE || lit == UNASSIGNED;
}

#endif
/*------------------------------------------------------------------------*/

static int
is_root_reason (Reason * r)
{
  return r == TOPLEVEL_REASON || r == DECISION_REASON || r == FORCED_REASON;
}

/*------------------------------------------------------------------------*/

static Clause *
reason2clause (Reason * reason)
{
  Clause *res;

  assert (!is_implication (reason));

  res = (Clause *) reason;

  return res;
}

/*------------------------------------------------------------------------*/

static Literal *
reason2premisse (Reason * reason)
{
  Literal *res;
  Word tmp;

  assert (is_implication (reason));

  /* Root reasons are implications but should never be transformed into
   * literals.
   */
  assert (!is_root_reason (reason));

  tmp = (Word) reason;
  tmp &= ~IMPLICATION_BIT;
  res = (Literal *) tmp;

  return res;
}

/*------------------------------------------------------------------------*/
#ifdef TRACE
/*------------------------------------------------------------------------*/

static const char *
reason2str (Reason * reason)
{
  Clause * clause;
  Literal * lit;

  if (reason == TOPLEVEL_REASON)
    return "top_level";

  if (reason == DECISION_REASON)
    return "decision";

  if (reason == FORCED_REASON)
    return "forced";

  assert (!is_root_reason (reason));

  if (is_implication (reason))
    {
      lit = reason2premisse (reason);
      sprintf (reason2str_buffer, "implication(%d)", lit2int (lit));
    }
  else
    {
      assert (reason);
      clause = reason2clause (reason);
      sprintf (reason2str_buffer, "clause(%d)", clause->idx);
    }

  return reason2str_buffer;
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

static void
check_reason (Reason * reason)
{
#ifdef CHECK
  Literal **p, **top_of_literals, *lit, *tmp;
  unsigned count, size;
  Clause *clause;

  if (is_root_reason(reason))
    return;

  if (is_implication (reason))
    {
      lit = reason2premisse (reason);
      tmp = deref (lit);
      assert (tmp == TRUE);
    }
  else
    {
      clause = reason2clause (reason);
      p = clause->literals;
      size = size_literals (p);
      top_of_literals = p + size;
      count = 0;
      while (p < top_of_literals)
	{
	  lit = *p++;
	  tmp = deref (lit);
	  if (tmp == FALSE)
	    count++;
	}
      assert (count + 1 == size);
    }
#endif
}

/*------------------------------------------------------------------------*/
#ifdef SIG
/*------------------------------------------------------------------------*/

static unsigned
level2sig (unsigned level)
{
  assert (sizeof (unsigned) >= 4);
  return 1 << (level & 31);
}

/*------------------------------------------------------------------------*/

static unsigned
lit2sig (Literal * lit)
{
  Variable * v;
  unsigned res;

  v = lit2var (lit);
  res = v->sig;

  return res;
}

/*------------------------------------------------------------------------*/

static unsigned
clause2sig (Clause * clause)
{
  return clause->sig;
}

/*------------------------------------------------------------------------*/

static unsigned
reason2sig (Reason * reason)
{
  if (is_root_reason (reason))
    return level2sig (decision_level);
  else if (is_implication (reason))
    return lit2sig (reason2premisse (reason));
  else
    return clause2sig (reason2clause (reason)); 
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

static void
assign (Literal * lit, Reason * reason)
{
  Variable *v;
  int sign;

  assert (!is_assigned (lit));

  check_reason (reason);

  INCSTAT (assignments, 1);

  v = lit2var (lit);
  sign = lit2sign (lit);
  v->assignment = sign2const (sign);
  v->reason = decision_level ? reason : TOPLEVEL_REASON;
  v->level = decision_level;
#ifdef SIG
  v->sig = reason2sig (v->reason);
#endif
  num_assigned++;
#ifdef STATS
  if (!decision_level)
    INCSTAT(top_level_assignments, 1);
#endif
  *top_of_trail++ = lit;

  dequeue_from_activity_queue (v);

  trace3 ("assign(%d,%s)", lit2int (lit), reason2str (reason));

#if 1
  {
    unsigned component = v->component - components;
    if (component != active_component)
      log2 ("component(%u)", component);
    active_component = component;
  }
#endif
}

/*------------------------------------------------------------------------*/

static void
unassign (Literal * lit)
{
  Variable *v;

  assert (is_assigned (lit));

  trace2 ("unassign(%d)", lit2int (lit));

  v = lit2var (lit);

  enqueue_on_activity_queue (v);

  assert (v->assignment == sign2const (lit2sign (lit)));
  v->assignment = 0;

  assert (num_assigned > 0);
  num_assigned--;
}

/*------------------------------------------------------------------------*/

static int
find_units (void)
{
  int res;

  res = count_stack(units) - old_count_units;
  assert (res >= 0);
#ifdef STATS
  log2 ("found %d new units", res);
#endif

  return res;
}

/*------------------------------------------------------------------------*/

static Literal *
is_unate (Variable * v)
{
  int pos, neg, idx;
  Literal *res;

  idx = sign2idx (1);
  pos = v->clauses[idx] || v->implications[idx];
  neg = v->clauses[!idx] || v->implications[!idx];

  if (pos && !neg)
    res = var2lit (v, 1);
  else if (!pos && neg)
    res = var2lit (v, -1);
  else
    res = 0;

  return res;
}

/*------------------------------------------------------------------------*/

static int
is_top_level_assigned_variable (Variable * v)
{
  return v->assignment != UNASSIGNED && !v->level;
}

/*------------------------------------------------------------------------*/

static int
find_unates (void)
{
  Literal *lit;
  Variable *v;
  int res;

  for (v = variables + 1; v < top_of_variables; v++)
    {
      if (v->unate || v->unit)
	continue;

      if (is_top_level_assigned_variable (v))
	continue;

      if (!(lit = is_unate (v)))
	continue;

      trace2 ("unate(%d)", lit2int (lit));
      push_stack (unates, lit);
      v->unate = 1;
    }

  res = count_stack (unates) - old_count_unates;
  assert (res >= 0);
#ifdef STATS
  log2 ("found %d new unates", res);
#endif

  return res;
}

/*------------------------------------------------------------------------*/

static Literal *
assign_units (void)
{
  Literal *res, **p, *lit, *tmp;

  res = TRUE;
  for (p = units + old_count_units; p < top_of_units; p++)
    {
      lit = *p;
      tmp = deref (lit);
      if (tmp == FALSE)
	res = FALSE;
      else if (tmp != TRUE)
	assign (lit, TOPLEVEL_REASON);
    }

  if (res == FALSE)
    conflicts++;

  old_count_units = count_stack (units);;

  return res;
}

/*------------------------------------------------------------------------*/

static void
assign_unates (void)
{
  Literal *lit, **p;

  for (p = unates + old_count_unates; p < top_of_unates; p++)
    {
      lit = *p;
      assert (deref (lit) != FALSE);
      if (!is_assigned (lit))
	assign (lit, TOPLEVEL_REASON);
    }

  old_count_unates = count_stack (unates);
}

/*------------------------------------------------------------------------*/

static int
all_assigned (void)
{
  return num_assigned == max_idx;
}

/*------------------------------------------------------------------------*/

static unsigned
length (void *anchor)
{
  unsigned res;
  void *p;

  res = 0;

  for (p = anchor; p; p = *(void **) p)
    res++;

  return res;
}

/*------------------------------------------------------------------------*/

static unsigned
hash_binary (unsigned size, unsigned prime, Literal * a, Literal * b)
{
  unsigned res;

  res = (unsigned) lit2int (a);
  res *= prime;
  res += (unsigned) lit2int (b);
  res &= size - 1;

  assert (res < size);

  return res;
}

/*------------------------------------------------------------------------*/

static unsigned
hash_binary_pos (unsigned size, Literal * a, Literal * b)
{
  return hash_binary (size, 2000000011, a, b);
}

/*------------------------------------------------------------------------*/

static unsigned
hash_binary_delta (unsigned size, Literal * a, Literal * b)
{
  unsigned res;

  res = hash_binary (size, 1000000007, a, b);

  if (!(res & 1))
    res++;

  assert (res < size && (res & 1));

  return res;
}

/*------------------------------------------------------------------------*/

static int
is_equal_binary (Binary * p, Literal * a, Literal * b)
{
  assert (p->first);

  return p->first == a && p->second == b;
}

/*------------------------------------------------------------------------*/

static Binary *
find_binary (Binary * table, unsigned size, Literal * a, Literal * b)
{
  unsigned h, d;
  Binary *res;

  h = hash_binary_pos (size, a, b);
  res = table + h;

  if (res->first && !is_equal_binary (res, a, b))
    {
      d = hash_binary_delta (size, a, b);

      do
	{
	  h += d;

	  if (h >= size)
	    {
	      h -= size;
	      assert (h < size);
	    }

	  res = table + h;
	}
      while (res->first && !is_equal_binary (res, a, b));
    }

  return res;
}

/*------------------------------------------------------------------------*/

static void
insert_binary (Binary * table, unsigned size, Literal * a, Literal * b)
{
  Binary *p;

  p = find_binary (table, size, a, b);

  if (p->first)
    {
      assert (p->first == a);
      assert (p->second == b);
    }
  else
    {
      assert (a);
      p->first = a;
      p->second = b;
    }

}

/*------------------------------------------------------------------------*/

static int
contains_binary (Binary * table, unsigned size, Literal * a, Literal * b)
{
  Binary *p;
  int res;

  p = find_binary (table, size, a, b);
  res = (p->first != 0);
  assert (!res || p->first == a);

  return res;
}

/*------------------------------------------------------------------------*/

static int
is_zombie (Variable * v)
{
  int i;

  for (i = 0; i <= 1; i++)
    if (v->implications[i] || v->clauses[i])
      return 0;

  return 1;
}

/*------------------------------------------------------------------------*/

static void
find_zombies (void)
{
  Variable *v;

  num_zombies = 0;
  num_non_zombies = 1;

  for (v = variables + 1; v < top_of_variables; v++)
    {
      if (is_zombie (v))
	{
	  v->zombie = 1;
	  trace2 ("dead(%d)", var2idx (v));
	  num_zombies++;
	}
      else
	{
	  v->zombie = 0;
	  trace2 ("live(%d)", var2idx (v));
	  num_non_zombies++;
	}
    }

  log2 ("found %u dead variables", num_zombies);
  log2 ("found %u live variables", num_non_zombies);
}

/*------------------------------------------------------------------------*/

static Binary *
new_binaries (unsigned *size_ptr)
{
  Literal *this, *other;
  unsigned size, sign_idx;
  int sign;
  Binary *binaries;
  Variable *v;
  size_t bytes;
  Link *p;

  size = 0;

  for (v = variables + 1; v < top_of_variables; v++)
    for (sign_idx = 0; sign_idx <= 1; sign_idx++)
      size += length (v->implications[sign_idx]);

  size = (3 * size) / 2 + 1;
  size = (1 << ldceil (size));

  bytes = sizeof (binaries[0]) * size;
  binaries = nanosat_malloc (bytes);
  memset (binaries, 0, bytes);

  for (v = variables + 1; v < top_of_variables; v++)
    {
      for (sign_idx = 0; sign_idx <= 1; sign_idx++)
	{
	  sign = idx2sign (sign_idx);
	  this = var2lit (v, sign);

	  for (p = v->implications[sign_idx]; p; p = cdr (p))
	    {
	      other = car (p);
	      insert_binary (binaries, size, this, other);
	    }
	}
    }

  *size_ptr = size;

  return binaries;
}

/*------------------------------------------------------------------------*/

static void
delete_binaries (Binary * binaries, unsigned size)
{
  nanosat_free (binaries, sizeof (binaries[0]) * size);
}

/*------------------------------------------------------------------------*/

static void
find_and_or_gates (Binary * binaries, unsigned size)
{
  Literal **q, **top_of_literals, *this, *other, *not_this, *not_other;
  unsigned sign_idx;
  int sign, dependent;
  Clause *clause;
  Variable *v;
  Link *p;

  for (v = variables + 1; v < top_of_variables; v++)
    {
      if (v->dependent)
	continue;

      dependent = 0;

      for (sign_idx = 0; !dependent && sign_idx <= 1; sign_idx++)
	{
	  sign = idx2sign (sign_idx);
	  this = var2lit (v, sign);
	  not_this = not_lit (this);

	  for (p = v->clauses[sign_idx]; !dependent && p; p = cdr (p))
	    {
	      clause = car (p);
	      q = clause->literals;
	      top_of_literals = q + size_literals (q);

	      dependent = 1;

	      while (dependent && q < top_of_literals)
		{
		  other = *q++;
		  if (this == other)
		    continue;

		  not_other = not_lit (other);

		  if (!contains_binary (binaries, size, not_this, not_other))
		    dependent = 0;
		}
	    }
	}

      if (dependent)
	{
	  v->dependent = 1;
	  trace2 ("dependent(%d) <and/or>", var2idx (v));
	  num_dependent++;
	}
    }
}

/*------------------------------------------------------------------------*/

static void
find_equalities (Binary * binaries, unsigned size)
{
#if 0				/* DOES NOT WORK */
  Literal *this, *other, *not_this, *not_other;
  int sign, dependent;
  unsigned sign_idx;
  Variable *v;
  Link *p;

  for (v = variables + 1; v < top_of_variables; v++)
    {
      if (v->dependent)
	continue;

      dependent = 0;

      for (sign_idx = 0; !dependent && sign_idx <= 1; sign_idx++)
	{
	  sign = idx2sign (sign_idx);
	  this = var2lit (v, sign);
	  not_this = not_lit (this);

	  for (p = v->implications[sign_idx]; !dependent && p; p = cdr (p))
	    {
	      other = car (p);
	      not_other = not_lit (other);

	      if (contains_binary (binaries, size, not_this, not_other))
		dependent = 1;
	    }
	}

      if (dependent)
	{
	  v->dependent = 1;
	  trace2 ("dependent(%d) <equality>", var2idx (v));
	  num_dependent++;
	}
    }
#endif
}

/*------------------------------------------------------------------------*/

static void
find_independent (void)
{
  unsigned num_free;
  Variable *v;

  num_independent = 0;

  for (v = variables + 1; v < top_of_variables; v++)
    {
      if (v->dependent)
	continue;

      trace2 ("independent(%d)", var2idx (v));
      num_independent++;
    }

#ifndef NDEBUG
  if (count_stack (variables))
    assert (num_dependent + num_independent + 1 == count_stack (variables));
#endif

  log2 ("found %u independent variables", num_independent);

  assert (num_independent >= num_zombies);
  num_free = num_independent - num_zombies;

  log3 ("found %u really free variables (%.1f%% of live)", 
        num_free, 
	num_non_zombies ? (100.0 * num_free) / num_non_zombies : 0.0);

  if (num_free <= 40)
    {
      enable_cmp_var_independent = 1;
      log1 ("focusing on those less than 40 independent variables");
    }
  else
  if (num_free * 10 < num_non_zombies)
    {
      enable_cmp_var_independent = 1;
      log1 ("focusing on those less than 10%% indepented variables");
    }
  else
    {
      enable_cmp_var_independent = 0;
      log1 ("handling independent and dependent variables in the same way");
    }
}

/*------------------------------------------------------------------------*/

static void
find_dependent (void)
{
  Binary *binaries;
  unsigned size;

  num_dependent = 0;

  binaries = new_binaries (&size);
  find_equalities (binaries, size);
  find_and_or_gates (binaries, size);
  delete_binaries (binaries, size);

  log2 ("found %u dependent variables", num_dependent);
}

/*------------------------------------------------------------------------*/

static Reason *
bcp_implications (void)
{
  Literal *lit, *other, *tmp;
  Reason *reason;
  Link *p;

  assert (unprocessed_implications <= top_of_trail);

  while (unprocessed_implications != top_of_trail)
    {
      lit = *unprocessed_implications++;
      reason = premisse2reason (lit);

      for (p = *lit2implicationsptr (not_lit (lit)); p; p = cdr (p))
	{
	  INCSTAT (implications, 1);

	  other = car (p);

	  tmp = deref (other);
	  if (tmp == TRUE)
	    continue;

	  if (tmp == FALSE)
	    return premisse2reason (other);

	  assign (other, reason);
	}
    }

  return 0;
}

/*------------------------------------------------------------------------*/

static Link *
dec_counters (Literal * lit)
{
  Link *p, *start;
  Clause *clause;

  start = *lit2clausesptr (lit);

  for (p = start; p; p = cdr (p))
    {
      INCSTAT (visits, 1);
      clause = car (p);
      assert (clause->connected);
      assert (clause->count > 0);
      clause->count--;
      clause->sum ^= (Word) lit;
#ifdef SIG
      clause->sig |= lit2sig (lit);
#endif
    }

  return start;
}

/*------------------------------------------------------------------------*/

static void
inc_counters (Literal * lit)
{
  Link *p, *start;
  Clause *clause;

  start = *lit2clausesptr (lit);

  for (p = start; p; p = cdr (p))
    {
      INCSTAT (visits, 1);
      clause = car (p);
      assert (clause->connected);
      clause->count++;
      assert (clause->count);
      clause->sum ^= (Word) lit;
    }
}

/*------------------------------------------------------------------------*/

static Reason *
bcp_clauses (void)
{
  Literal *lit, *tmp, *unit;
  Reason *reason, *conflict;
  Link *p, *start;
  Clause *clause;

  assert (unprocessed_clauses <= top_of_trail);

  while (unprocessed_clauses != top_of_trail)
    {
      lit = not_lit (*unprocessed_clauses++);

      start = dec_counters (lit);

      for (p = start; p; p = cdr (p))
	{
	  clause = car (p);
	  assert (clause->count);

	  if (clause->count != 1)
	    continue;

	  unit = (Literal *) clause->sum;
	  tmp = deref (unit);

	  if (tmp != TRUE)
	    {
	      reason = clause2reason (clause);
	      if (tmp == FALSE)
		return reason;

	      assign (unit, reason);
	      if ((conflict = bcp_implications ()))
		return conflict;
	    }
	}
    }

  return 0;
}

/*------------------------------------------------------------------------*/
#ifdef CHECK
/*------------------------------------------------------------------------*/

static void
check_clause (Clause * clause)
{
  Literal **p, **top_of_literals, *tmp, *lit;
  unsigned count;
  Word sum;

  if (!clause->connected)
    return;

  sum = 0;

  p = clause->literals;
  top_of_literals = p + size_literals (p);
  count = 0;

  while (p < top_of_literals)
    {
      lit = *p++;
      tmp = deref (lit);
      if (tmp == FALSE)
	continue;

      count++;
      sum ^= (Word) lit;
    }

  assert (clause->sum == sum);
  assert (clause->count == count);
}

/*------------------------------------------------------------------------*/

static void
check_clauses (void)
{
  Clause **p;

  assert (2 * num_binary_clauses == num_implications);

  for (p = clauses; p < top_of_clauses; p++)
    check_clause (*p);
}

/*------------------------------------------------------------------------*/

static int
clause_is_satisfied (Clause * clause)
{
  Literal **p, **top_of_literals;

  p = clause->literals;
  top_of_literals = p + size_literals (p);

  while (p < top_of_literals)
    if (deref (*p++) == TRUE)
      return 1;

  return 0;
}

/*------------------------------------------------------------------------*/

static void
check_frame (Frame * frame)
{
  Clause **p, **satisfied, **top_of_satisfied;

  satisfied = clauses + frame->satisfied;
  top_of_satisfied = clauses + frame->top_of_satisfied;

  assert (clauses <= satisfied);
  assert (satisfied <= top_of_satisfied);
  assert (top_of_satisfied <= top_of_clauses);

  for (p = satisfied; p < top_of_satisfied; p++)
    assert (clause_is_satisfied (*p));
}

/*------------------------------------------------------------------------*/

static void
check_frames (void)
{
  Frame *p;

  for (p = control; p < top_of_control; p++)
    check_frame (p);
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

static void
check_invariant (void)
{
#ifdef CHECK
  check_clauses ();
  check_frames ();
#endif
}

/*------------------------------------------------------------------------*/

static Reason *
bcp (void)
{
  Reason *conflict;

  conflict = bcp_implications ();
  if (!conflict)
    conflict = bcp_clauses ();

  if (conflict)
    {
      conflicts++;
      trace2 ("conflict(%s)", reason2str (conflict));
    }
  else
    check_invariant ();

  return conflict;
}

/*------------------------------------------------------------------------*/

static int
is_top_level_satisfied_literal (Literal * lit)
{
  return deref (lit) == TRUE && !lit2var (lit)->level;
}

/*------------------------------------------------------------------------*/

static int
is_top_level_satisfied_clause (Clause * clause)
{
  Literal **p, **top_of_literals;

  p = clause->literals;
  top_of_literals = p + size_literals (p);

  while (p < top_of_literals)
    if (is_top_level_satisfied_literal (*p++))
      return 1;

  return 0;
}

/*------------------------------------------------------------------------*/

static int
is_relevant_clause (Clause * clause)
{
  unsigned num_unassigned, size, remaining;
  Literal **p, **top_of_literals;

  p = clause->literals;
  size = size_literals (p);
  if (size <= MAX_RELEVANCE)
    return 1;

  remaining = size - MAX_RELEVANCE;
  top_of_literals = p + size;
  num_unassigned = 0;

  while (p < top_of_literals)
    {
      if (is_assigned (*p++))
	{
	  remaining--;
	  if(!remaining)
	    return 1;
	}
      else
	num_unassigned++;
    }

  return num_unassigned <= MAX_RELEVANCE;
}

/*------------------------------------------------------------------------*/

static unsigned
size_clause (Clause * clause)
{
  return size_literals (clause->literals);
}

/*------------------------------------------------------------------------*/

static int
always_keep_clause (Clause * clause)
{
  return size_clause (clause) <= MAX_ALWAYS_KEPT_CLAUSE_SIZE;
}

/*------------------------------------------------------------------------*/

static int
cmp_ptr (const void * a, const void * b)
{
  if (a < b)
    return -1;

  if (a > b)
    return 1;

  return 0;
}

/*------------------------------------------------------------------------*/

static int
cmp_unsigned (unsigned a, unsigned b)
{
  if (a < b)
    return -1;

  if (a > b)
    return 1;

  return 0;
}

/*------------------------------------------------------------------------*/

static int
cmp_candidates (const void *p, const void *q)
{
  const GCCandidate *a, *b;
  unsigned m, n;
  Clause *c, *d;
  int res;

  a = p;
  b = q;

  c = a->clause;
  d = b->clause;

  m = size_clause (c);
  n = size_clause (d);

  if (2 * m <= n)
    return 1;

  if (m >= 2 * n)
    return -1;

  if ((res = cmp_unsigned (c->activity, d->activity)))
    return res;

  res = -cmp_unsigned (a->age, b->age);

  return res;
}

/*------------------------------------------------------------------------*/
#if defined(LOG) || defined(STATS)
/*------------------------------------------------------------------------*/

static double
connection_efficiency (void)
{
  double res;

  if (sum_connected_clauses)
    {
      assert (sum_multiple_used_connected_clauses <= sum_connected_clauses);

      res = sum_multiple_used_connected_clauses;
      res /= sum_connected_clauses;
    }
  else
    res = 0;

  return res;
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

static double
learning_efficiency (void)
{
  double res;

  if (sum_learned_clauses)
    {
      assert (sum_multiple_used_learned_clauses <= sum_learned_clauses);

      res = sum_multiple_used_learned_clauses;
      res /= sum_learned_clauses;
    }
  else
    res = 0;

  return res;
}

/*------------------------------------------------------------------------*/

static int
is_high_learning_efficiency (double efficiency)
{
  return efficiency > 0.4;
}

/*------------------------------------------------------------------------*/

static unsigned
gc_goal (void)
{
  unsigned res;
#if 1
  double efficiency;

  efficiency = learning_efficiency ();

  log4 ("gc(%.0f) %s learning efficiency %.1f%%", 
        (double) gcs,
	is_high_learning_efficiency (efficiency) ? "high" : "low", 
	100.0 * efficiency);

  if (sum_learned_clauses >= 500)
    {
      if (is_high_learning_efficiency (efficiency))
	{
	  dump_factor *= 0.99;
	  if (dump_factor < 0.2)
	    dump_factor = 0.2;
	}
      else
	{
	  dump_factor *= 1.02;		/* should adapt faster! */
	  if (dump_factor > 0.9)
	    dump_factor = 0.9;
	}
    }

#endif
  res = num_learned_clauses;
  res *= dump_factor;

  log4 ("gc(%.0f) goal is to collect %u learned clauses = %.1f%%",
        (double) gcs, res, 100.0 * dump_factor);

  return res;
}

/*------------------------------------------------------------------------*/

static void
gc_mark (void)
{
  unsigned goal_collected_learned, num_collected_learned, num_forced_learned;
  declare_stack (/* automatic */, GCCandidate, candidates);
  unsigned num_candidates, num_collected, i, num_forced;
  GCCandidate candidate;
  Clause **p, *clause;
  unsigned num_kept;
  Reason * reason;
  Literal **q;
  Variable *v;
#ifdef LOG
  unsigned old_num_clauses = count_stack (clauses);
#endif
#ifdef TRACE
  const char * collection_msg;
#endif

  /* First mark all clauses occurring on the trail as to keep.
   */
  num_kept = 0;
  for (q = trail; q < top_of_trail; q++)
    {
      v = lit2var (*q);
      if (v->assignment == UNASSIGNED)
	continue;
	
      reason = v->reason;
      if (is_implication (reason))
	continue;

      clause = reason2clause (reason);
      assert (!clause->keep);

      clause->keep = 1;
      trace3 ("keep(%d): clause(%d) is on trail", num_kept, clause->idx);

      num_kept++;
    }

  /* Then determine the necessarily to be kept and necessarily to be
   * garbage collected clauses.
   */
  num_collected = 0;
  num_collected_learned = 0;
  for (p = clauses; p < top_of_clauses; p++)
    {
      clause = *p;
      if (clause->keep)
	continue;

#ifdef TRACE
      collection_msg = 0;
#endif
      if (is_top_level_satisfied_clause (clause))
	{
	  clause->garbage = 1;
	  INCSTAT(top_level_satisfied_clauses, 1);
#ifdef TRACE
	  collection_msg = "top level satisfied";
#endif
	}
      else if (clause->external || always_keep_clause (clause))
	{
	  clause->keep = 1;
	}
      else if (!is_relevant_clause (clause))
	{
	  clause->garbage = 1;
	  INCSTAT(irrelevant, 1);
#ifdef TRACE
	  collection_msg = "irrelevant";
#endif
	}
      else if (p + NUM_STICKY_RECENTLY_LEARNED > top_of_clauses)
	{
	  clause->keep = 1;
#ifdef TRACE
	  collection_msg = "recently learned";
#endif
	}

      if (clause->garbage)
	{
	  num_collected++;

	  if (!clause->external)
	    num_collected_learned++;
	}

#ifdef TRACE
      assert (!collection_msg == !clause->garbage);

      if (collection_msg)
	trace4 ("collect(%d): clause(%d) is %s",
		num_collected, clause->idx, collection_msg);
#endif
    }

  num_forced = num_collected;
  num_forced_learned = num_collected_learned;

  goal_collected_learned = gc_goal ();

  init_stack (candidates);

  p = clauses;
  if (valid_first_learned_clause_idx)
    p += first_learned_clause_idx;

  while (p < top_of_clauses)
    {
      clause = *p++;
      if (clause->external)
	continue;

      candidate.clause = clause;
      candidate.age = top_of_clauses - p;
      push_stack (candidates, candidate);
    }

  num_candidates = count_stack (candidates);

  qsort (candidates, num_candidates, sizeof (candidates[0]), cmp_candidates);

  for (i = 0; i < num_candidates && i < goal_collected_learned; i++)
    {
      clause = candidates[i].clause;
      if (clause->keep || clause->garbage)
	continue;

      clause->garbage = 1;
      trace4 ("collect(%d): learned clause(%d) is candidate(%d)",
	      num_collected, clause->idx, i);

      num_collected_learned++;
      num_collected++;
    }

  release_stack (candidates);

  INCSTAT(collected, num_collected);

  log5 ("gc(%.0f) collected %u learned clauses = %.1f%% (%.1f%% forced)",
        (double) gcs, num_collected_learned,
	num_candidates ? 
	  (100.0 * num_collected_learned) / num_candidates : 100.0,
	num_collected_learned ?
	  (100.0 * num_forced_learned) / num_collected_learned : 100.0);

  log5 ("gc(%.0f) collected %u clauses = %.1f%% (%.1f%% forced)",
	(double) gcs, num_collected,
	(100.0 * num_collected) / old_num_clauses,
	num_collected ? (100.0 * num_forced) / num_collected : 100.0);
}

/*------------------------------------------------------------------------*/

static void
move_literals (void)
{
  Clause **p, *clause;

  for (p = clauses; p < top_of_clauses; p++)
    {
      clause = *p;
      if (clause->garbage)
	continue;

      /* TODO resize clauses with top_level assigned literals to false.
       */
      clause->literals = realloc_literals (clause->literals);
    }
}

/*------------------------------------------------------------------------*/

static void
recycle_clause (Clause * clause)
{
  if (!clause->external)
    {
      assert (num_learned_clauses > 0);
      num_learned_clauses--;
    }

  assert (num_clauses > 0);
  num_clauses--;
}

/*------------------------------------------------------------------------*/

static void
move_clause (Clause * old_clause)
{
  Clause *new_clause;

  assert (!old_clause->garbage);

  new_clause = new (sizeof (*new_clause));
  memcpy (new_clause, old_clause, sizeof (*new_clause));
  new_clause->keep = 0;
  old_clause->literals = (Literal **) new_clause;
  old_clause->count = count_stack (clauses);
  *top_of_clauses++ = new_clause;
}

/*------------------------------------------------------------------------*/

static void
move_clauses (void)
{
  Clause **p, **old_top_of_clauses, *old_clause;

  old_top_of_clauses = top_of_clauses;
  top_of_clauses = clauses;

  for (p = clauses; p < old_top_of_clauses; p++)
    {
      old_clause = *p;

      if (old_clause->garbage)
	recycle_clause (old_clause);
      else
	move_clause (old_clause);
    }

  assert (top_of_clauses <= old_top_of_clauses);
}

/*------------------------------------------------------------------------*/

static void
fix_first_learned (void)
{
  Clause **p, *clause;

  valid_first_learned_clause_idx = 0;

  for (p = clauses; p < top_of_clauses; p++)
    {
      clause = *p;
      if (clause->external)
	continue;

      first_learned_clause_idx = p - clauses;
      valid_first_learned_clause_idx = 1;
      break;
    }
}

/*------------------------------------------------------------------------*/

static void *
reverse_list (void *anchor)
{
  void *next, *prev, *this;

  prev = 0;

  for (this = anchor; this; this = next)
    {
      next = *(void **) this;
      *(void **) this = prev;
      prev = this;
    }

  return prev;
}

/*------------------------------------------------------------------------*/

static void
dec_num_occurences (Literal * lit)
{
  unsigned sign_idx = sign2idx (lit2sign (lit));
  Variable *v = lit2var (lit);

  assert (v->num_occurences[sign_idx] > 0);
  v->num_occurences[sign_idx] -= 1;
}

/*------------------------------------------------------------------------*/

static void
dec_num_occurences_of_clause (Clause * clause)
{
  Literal **p, **top_of_literals;

  p = clause->literals;
  top_of_literals = p + size_literals (p);

  while (p < top_of_literals)
    dec_num_occurences (*p++);
}

/*------------------------------------------------------------------------*/

static void
recycle_implication (Variable * v, int sign_idx, Literal * conclusion)
{
#ifdef TRACE
  Literal * this = var2lit (v, idx2sign(sign_idx));
  trace3 ("delete(implication(%d,%d))",
          lit2int (not_lit(this)), lit2int (conclusion));
#endif
  assert (num_implications > 0);
  num_implications--;

  if (!(num_implications & 1))
    {
      assert (num_binary_clauses > 0);
      num_binary_clauses--;
    }
}

/*------------------------------------------------------------------------*/

static void
move_occurences_in_implications (Variable * v, int sign_idx)
{
  Link *p, *old_implications, *new_implications;
  Literal *lit;

  new_implications = 0;
  old_implications = reverse_list (v->implications[sign_idx]);

  for (p = old_implications; p; p = cdr (p))
    {
      lit = car (p);
      if (is_top_level_assigned_variable (lit2var (lit)))
	recycle_implication (v, sign_idx, lit);
      else
	new_implications = cons (lit, new_implications);
    }

  v->implications[sign_idx] = new_implications;
}

/*------------------------------------------------------------------------*/

static void
move_occurences_in_clauses (Variable * v, int sign_idx)
{
  Link *p, *new_clauses, *old_clauses;
  Clause *old_clause, *new_clause;

  new_clauses = 0;
  old_clauses = reverse_list (v->clauses[sign_idx]);

  for (p = old_clauses; p; p = cdr (p))
    {
      old_clause = car (p);
      if (old_clause->garbage)
	continue;

      new_clause = (Clause *) old_clause->literals;
      new_clauses = cons (new_clause, new_clauses);
    }
  v->clauses[sign_idx] = new_clauses;
}

/*------------------------------------------------------------------------*/

static void
flush_occurences (Variable * v)
{
  int sign_idx;
  Link * p;

  for (sign_idx = 0; sign_idx <= 1; sign_idx++)
    {
      for (p = v->implications[sign_idx]; p; p = cdr(p))
	recycle_implication (v, sign_idx, car(p));

      v->implications[sign_idx] = 0;
      v->clauses[sign_idx] = 0;
    }
}

/*------------------------------------------------------------------------*/

static void
move_occurrences (void)
{
  int sign_idx;
  Variable *v;
#ifdef LOG
  unsigned old_num_implications = num_implications;
#endif

  for (v = variables + 1; v < top_of_variables; v++)
    {
      if (!is_top_level_assigned_variable (v))
	{
	  for (sign_idx = 0; sign_idx <= 1; sign_idx++)
	    {
	      move_occurences_in_implications (v, sign_idx);
	      move_occurences_in_clauses (v, sign_idx);
	    }
	}
      else
	flush_occurences (v);
    }

  log3 ("gc(%.0f) collected %u implications (all forced)", 
        (double) gcs, old_num_implications - num_implications);
}

/*------------------------------------------------------------------------*/

static void
gc_clauses (void)
{
  Clause **p, *clause;

  for (p = clauses; p < top_of_clauses; p++)
    {
      clause = *p;
      if (clause->garbage)
	dec_num_occurences_of_clause (clause);
    }
}

/*------------------------------------------------------------------------*/

static void
fix_reasons (void)
{
  Clause *old_clause, *new_clause;
  Reason *reason;
  Variable *v;

  for (v = variables + 1; v < top_of_variables; v++)
    {
      if (v->assignment == UNASSIGNED)
	continue;
      
      reason = v->reason;
      if (is_implication (reason))
	continue;

      old_clause = reason2clause (reason);
      if (old_clause->garbage)
	{
	  assert (is_top_level_satisfied_clause (old_clause));
	  continue;
	}

      new_clause = (Clause *) old_clause->literals;
      v->reason = clause2reason (new_clause);
    }
}

/*------------------------------------------------------------------------*/

static void
set_default_satisfied_interval (Frame * frame)
{
  frame->top_of_satisfied = count_stack (clauses);
  frame->satisfied = frame->top_of_satisfied;
}

/*------------------------------------------------------------------------*/

static void
fix_frame (Frame * frame)
{
  set_default_satisfied_interval (frame);
}

/*------------------------------------------------------------------------*/

static void
fix_frames (void)
{
  Frame * p;

  for (p = control; p < top_of_control; p++)
    fix_frame (p);
}

/*------------------------------------------------------------------------*/

static void
gc_move (void)
{
#ifdef LOG
  unsigned bytes_before_collection = dynamic_allocated_bytes;
  unsigned delta_recycled_bytes;
#endif
  Chunk *old_chunks;
  
  old_chunks = chunks;		/* 1. remember src chunks for deletion */
  chunks = 0;			/* 2. initialize target chunks list */
  new_chunk (allocated_mem);	/* 3. allocate one contiguous target chunk */

  move_literals ();
  move_clauses ();
  fix_first_learned ();
  move_occurrences ();
  fix_reasons ();
  fix_frames ();

  delete_chunks (old_chunks);	/* delete old src chunks */
#ifdef LOG
  assert (dynamic_allocated_bytes <= bytes_before_collection);
  delta_recycled_bytes = bytes_before_collection - dynamic_allocated_bytes;
  assert (bytes_before_collection);
  log4 ("gc(%.0f) collected %.1f MB = %.1f%%",
	(double) gcs, delta_recycled_bytes / (double) (1 << 20),
	(100.0 * delta_recycled_bytes) / bytes_before_collection);
#endif
}

/*------------------------------------------------------------------------*/
/* A moving garbage collector.
 */
static void
gc (int forced)
{
  gcs++;
  last_gc_conflicts = conflicts;

  if (!count_stack (clauses))
    log2 ("gc(%.0f) skipped since no clauses left", (double) gcs);
  else if (!forced && last_restart_conflicts == conflicts)
    log2 ("gc(%.0f) skipped without new conflict after restart",
	  (double) gcs);
  else
    {
      gc_mark ();
      gc_clauses ();
      gc_move ();
    }
}

/*------------------------------------------------------------------------*/

static double
get_time (void)
{
  struct rusage u;
  double res;

  res = 0;

  if (!getrusage (RUSAGE_SELF, &u))
    {
      res += u.ru_utime.tv_sec + 1e-6 * u.ru_utime.tv_usec;
      res += u.ru_stime.tv_sec + 1e-6 * u.ru_stime.tv_usec;
    }

  return res;
}

/*------------------------------------------------------------------------*/

static double
delta_time (void)
{
  double res;

  res = get_time () - entered_time;
  res = (res < 0) ? 0 : res;

  return res;
}

/*------------------------------------------------------------------------*/

static double
time_spent_in_library (void)
{
  return delta_time() + seconds;
}

/*------------------------------------------------------------------------*/

static void
untrail (Literal ** new_top_of_trail)
{
  while (top_of_trail > new_top_of_trail)
    unassign (*--top_of_trail);

  unprocessed_implications = top_of_trail;

  while (unprocessed_clauses > top_of_trail)
    inc_counters (not_lit (*--unprocessed_clauses));
}

/*------------------------------------------------------------------------*/

static void
adjust (Reason * conflict)
{
  Literal **p, *lit, **top_of_literals;
  int new_level;
  Clause *clause;
  Variable *v;

  if (is_implication (conflict))
    {
#ifndef NDEBUG
      lit = reason2premisse (conflict);
      v = lit2var (lit);
      assert (decision_level == v->level);
#endif
    }
  else
    {
      clause = reason2clause (conflict);
#ifdef NDEBUG
      if (!clause->connected)
#endif
	{
	  clause = reason2clause (conflict);
	  p = clause->literals;
	  top_of_literals = p + size_literals (p);

	  new_level = 0;

	  while (p < top_of_literals)
	    {
	      lit = *p++;
	      v = lit2var (lit);
	      if (v->level > new_level)
		new_level = v->level;
	    }

	  if (new_level < decision_level)
	    {
	      assert (!clause->connected);
	      assert (!clause->external);
	      trace2 ("adjust(%d)", new_level - decision_level);
	      INCSTAT (adjustments, 1);
	      INCSTAT (adjusted_levels, decision_level - new_level);
	      decision_level = new_level;
	    }
	}
    }

}

/*------------------------------------------------------------------------*/

static void
expand_literal (Literal * lit)
{
  Variable *v;

  assert (is_assigned (lit));

  v = lit2var (lit);
  if (v->visited)
    return;

  v->visited = 1;
  push_stack (visited, v);

  inc_variable_activity (v);

  assert (v->level <= decision_level);
  if (v->level < decision_level)
    push_lit (deref (lit) == TRUE ? not_lit (lit) : lit);
  else
    num_open_paths++;
}

/*------------------------------------------------------------------------*/

static void
expand_clause (Clause * clause)
{
  Literal **p, **literals, **top_of_literals;

  inc_clause_activity (clause);
  inc_clause_usage (clause);

  literals = clause->literals;
  top_of_literals = literals + size_literals (literals);

  for (p = literals; p < top_of_literals; p++)
    expand_literal (*p);
}

/*------------------------------------------------------------------------*/

static void
check_exists_implication (Literal * premisse, Literal * conclusion)
{
#ifdef CHECK
  int found = 0;
  Link *p;

  for (p = *lit2implicationsptr (not_lit (premisse)); p; p = cdr (p))
    if (car (p) == conclusion)
      found = 1;

  assert (found);
#endif
}

/*------------------------------------------------------------------------*/

static void
expand_implication (Reason * reason)
{
  Literal *premisse, *conclusion;

  assert (is_implication (reason));
  assert (unprocessed_implications > trail);

  conclusion = reason2premisse (reason);
  premisse = unprocessed_implications[-1];

  assert (deref (premisse) == TRUE);
  assert (deref (conclusion) == FALSE);
  check_exists_implication (premisse, conclusion);

  expand_literal (conclusion);
  expand_literal (premisse);
}

/*------------------------------------------------------------------------*/

static void
expand (Reason * reason)
{
  if (is_implication (reason))
    expand_literal (reason2premisse (reason));
  else
    expand_clause (reason2clause (reason));
}

/*------------------------------------------------------------------------*/

static void
expand_conflict (Reason * reason)
{
  if (is_implication (reason))
    expand_implication (reason);
  else
    expand_clause (reason2clause (reason));
}

/*------------------------------------------------------------------------*/

static void
reset_visited (void)
{
  Variable **p;

  for (p = visited; p < top_of_visited; p++)
    (*p)->visited = 0;

  top_of_visited = visited;
}

/*------------------------------------------------------------------------*/
#ifdef SIG
/*------------------------------------------------------------------------*/

static unsigned
sig_length (unsigned sig)
{
  unsigned res, i;

  res = 0;
  for (i = 0; i < 31; i++)
    if (sig & (1 << i))
      res++;

  res *= (decision_level + 31) / 32;

  return res;
}

/*------------------------------------------------------------------------*/

static void
analyze_sig (Reason * conflict)
{
  Literal ** p, * lit, ** top_of_literals;
  unsigned sig, len;
  Clause * clause;

  if (is_implication (conflict))
    {
      sig = lit2sign (reason2premisse (conflict));
      assert (unprocessed_implications > trail);
      sig += lit2sign (unprocessed_implications[-1]);
    }
  else
    {
      clause = reason2clause (conflict);

      sig = 0;
      p = clause->literals;
      top_of_literals = p + size_literals (p);

      while (p < top_of_literals)
	{
	  lit = *p++;
	  sig += lit2sig (lit);
	}
    }

  len = sig_length (sig);
  trace3 ("signature 0x%08x of length %u", sig, len);
  sum_learned_sigs++;
  sum_length_learned_sigs += len;
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

static void
analyze (Reason * conflict)
{
  Literal **p, *lit;
  Variable *v;

#ifdef SIG
  analyze_sig (conflict);
#endif
  assert (!num_open_paths);
  assert (!count_stack (pushed_literals));

  expand_conflict (conflict);
  p = top_of_trail;
  lit = FALSE;
  v = 0;

  do
    {
      lit = *--p;
      v = lit2var (lit);
      if (!v->visited)
	continue;

      if (--num_open_paths)
	{
	  if (!is_root_reason (v->reason))
	    expand (v->reason);
	}
      else
	{
	  inc_variable_activity (v);
	  push_lit (not_lit (lit));
	}
    }
  while (num_open_paths);

  assert (v);
  assert (lit != FALSE);

  if (!is_root_reason (v->reason))
    {
      INCSTAT (uips, 1);
      trace2 ("uip(%d)", lit2int (lit));
    }

  reset_visited ();
}

/*------------------------------------------------------------------------*/

static unsigned
jump (void)
{
  unsigned level, res;
  Literal **p;
  int delta;

  res = 0;
  for (p = pushed_literals; p < top_of_pushed_literals; p++)
    {
      level = lit2var (*p)->level;
      if (level != decision_level && level > res)
	res = level;
    }

  delta = res - decision_level;
  trace2 ("backtrack(%d)", delta);
  INCSTAT(backtracks, 1);

  if (decision_level)
    {
      assert (delta <= -1);
      INCSTAT(skipped_backtracks, 1 - delta);
    }

  return res;
}

/*------------------------------------------------------------------------*/

static void
undo (unsigned new_decision_level)
{
  Literal **new_top_of_trail;

  top_of_control = control + new_decision_level + 1;
  new_top_of_trail = top_of_control->old_top_of_trail;
  decision_level = new_decision_level;
  untrail (new_top_of_trail);
  check_invariant ();
}

/*------------------------------------------------------------------------*/

static Literal *
drive (void)
{
  Literal *res;

  /* DO NOT CHANGE THE NEXT LINE THIS WITHOUT FIXING 'learn()'
   */
  res = count_stack (pushed_literals) ? top_of_pushed_literals[-1] : FALSE;
  trace2 ("drive(%d)", lit2int (res));

  return res;
}

/*------------------------------------------------------------------------*/
#ifdef CHECK
#include <sys/types.h>
#include <sys/wait.h>
#endif

static void
check_learned_is_implied (void)
{
#ifdef CHECK
#if 0
  FILE *file = fopen ("/tmp/check-nanosat.cnf", "w");
  int system_res;
  Literal **p;
  dump (file, 0);
  for (p = pushed_literals; p < top_of_pushed_literals; p++)
    fprintf (file, "%d 0\n", lit2int (not_lit (*p)));
  fclose (file);
  system_res = system ("exec funex "
		       "/tmp/check-nanosat.cnf "
		       "--decision-limit=0 "
		       "--no-header " "-o /tmp/check-nanosat.log");
  system_res = WEXITSTATUS (system_res);
  assert (system_res == 20);
#endif
#endif
}

/*------------------------------------------------------------------------*/

static Reason *
learn (void)
{
  Literal * last_pushed_literal;
  unsigned size;
  Reason *res;

  check_learned_is_implied ();

  size = count_stack (pushed_literals);
  last_pushed_literal = size ? pushed_literals[0] : UNASSIGNED;

  push_learned_clause ();

  if (size < 2)
    res = TOPLEVEL_REASON;
  else if (size == 2)
    res = premisse2reason (not_lit (last_pushed_literal));
  else if (size > MAX_LEARNED_CLAUSE_SIZE)
    res = FORCED_REASON;
  else
    res = clause2reason (top_of_clauses[-1]);

  check_invariant ();

  return res;
}

/*------------------------------------------------------------------------*/

static void
next_iteration (void)
{
  /* Delay restart.
   */
  last_restart_conflicts = conflicts;

  log3 ("iteration(%u) at %.0f conflicts", iterations, (double)conflicts);

#if 0
  {
    char buf[200];
    FILE * out;

    sprintf (buf, "/tmp/nanosat%u.cnf", iterations);
    out = fopen (buf, "w");
    dump (out, 1);
    fclose (out);
    sprintf (buf, 
      "cnf2neato.awk /tmp/nanosat%u.cnf | neato -Tps > /tmp/nanosat%u.ps;"
      "gv -scale -2 /tmp/nanosat%u.ps &", 
      iterations, iterations, iterations);
    system (buf);
  }
#endif

  iterations++;
}

/*------------------------------------------------------------------------*/

static int
backtrack (Reason * conflict)
{
  unsigned new_decision_level;
  int top_level_reached;

  Literal *forced_lit;
  Reason *reason;

  adjust (conflict);
  top_level_reached = !decision_level;
  analyze (conflict);
  new_decision_level = jump ();
  undo (new_decision_level);

  forced_lit = 0;
  if (!top_level_reached)
    forced_lit = drive ();

  reason = learn ();

  check_invariant ();

  if (top_level_reached)
    return 0;

  if (!new_decision_level)
    next_iteration();

  assign (forced_lit, reason);

  return 1;
}

/*------------------------------------------------------------------------*/
#if 0

static Literal *
decide_static (void)
{
  Literal *res, *tmp;
  Variable *p;

  res = 0;

  for (p = variables + 1; !res; p++)
    {
      assert (p < top_of_variables);

      tmp = var2lit (p, 1);
      if (!is_assigned (tmp))
	res = tmp;
    }

  return res;
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

static Literal *
decide_phase (Variable * v)
{
  int sign, sign_idx;
  Literal *res;

  if (v->num_occurences[0] == v->num_occurences[1])
    sign_idx = v->last_active_phase;
  else
    sign_idx = (v->num_occurences[1] > v->num_occurences[0]);

  v->last_active_phase = sign_idx ? 1 : 0;

  sign = idx2sign (sign_idx);
  res = var2lit (v, sign);

  return res;
}

/*------------------------------------------------------------------------*/
#if 0
/*------------------------------------------------------------------------*/

static unsigned
max_unsigned (unsigned a, unsigned b)
{
  return (a < b) ? b : a;
}

/*------------------------------------------------------------------------*/

static unsigned
max_occurence (Variable * v)
{
  return max_unsigned (v->num_occurences[0], v->num_occurences[1]);
}

/*------------------------------------------------------------------------*/

static int
cmp_var_occurence (Variable * u, Variable * v)
{
  return cmp_unsigned (sum_occurence (u), sum_occurence (v));
}

/*------------------------------------------------------------------------*/

static unsigned
sum_occurence (Variable * v)
{
  return v->num_occurences[0] + v->num_occurences[1];
}

/*------------------------------------------------------------------------*/

static int
cmp_var_idx (Variable * u, Variable * v)
{
  return cmp_unsigned (var2idx(u), var2idx (v));
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

static int
cmp_var_independence (Variable * u, Variable * v)
{
  if (u->dependent && !v->dependent)
    return -1;

  if (!u->dependent && v->dependent)
    return 1;

  return 0;
}

/*------------------------------------------------------------------------*/

static int
cmp_var_order (Variable * u, Variable * v)
{
  return -cmp_unsigned (u->order, v->order);
}

/*------------------------------------------------------------------------*/

static int
cmp_var_component (Variable * u, Variable *v)
{
  assert (components <= u->component && u->component < top_of_components);
  assert (components <= v->component && v->component < top_of_components);

  return -cmp_ptr (u->component, v->component);
}

/*------------------------------------------------------------------------*/

static int
cmp_var_activity (Variable * u, Variable * v)
{
  unsigned a, b;

  a = u->activity;
  b = v->activity;

  return cmp_unsigned (a, b);
}

/*------------------------------------------------------------------------*/

static int
cmp_var (Variable * u, Variable * v)
{
  int res;

  if ((res = cmp_var_component (u, v)))
    return res;

  if ((res = cmp_var_activity (u, v)))
    return res;

  if (enable_cmp_var_independent && (res = cmp_var_independence (u, v)))
    return res;

  #if 0
  if ((res = cmp_var_occurence (u, v)))
    return res;
  #endif

  return cmp_var_order (u, v);
}

/*------------------------------------------------------------------------*/

static Literal *
decide_activity (void)
{
  ActivityAnchor *anchor;
  Variable *v, * best;
  Literal *res;
  int i;

  INCSTAT(decisions_based_on_activity,1);

  best = 0;

  for (;;)
    {
      anchor = activity_queue[max_ld_activity];
      v = anchor->first;

      for (i = 0; v && i < 100; i++)
	{
	  assert (v->assignment == UNASSIGNED);

	  if (!best || cmp_var (best, v) > 0)
	    best = v;

	  v = v->activity_chain.next;
	}

      if (best)
	break;

      assert (max_ld_activity > 0);
      max_ld_activity--;
    }

  assert (best);

  res = decide_phase (best);

  return res;
}

/*------------------------------------------------------------------------*/

static Frame *
current_frame (void)
{
  assert (count_stack (control));
  return top_of_control - 1;
}

/*------------------------------------------------------------------------*/

static Literal *
decide_connection (Reason ** reason_ptr)
{
  Literal **p, **top_of_literals, *lit, *tmp, *best_lit;
  unsigned old_satisfied, old_top_of_satisfied;
  Variable *best, *v;
  Reason *reason;
  Clause *clause;
  Frame *frame;
  int count;

  if (!valid_first_learned_clause_idx)
    return decide_activity ();

  frame = current_frame ();
  old_satisfied = frame->satisfied;
  old_top_of_satisfied = frame->top_of_satisfied;
  frame->top_of_satisfied = frame->satisfied = count_stack (clauses);

  for (;;)
    {
      if (frame->satisfied == old_top_of_satisfied)
	{
	  assert (frame->satisfied >= old_satisfied);
	  INCSTAT(hits, frame->satisfied - old_satisfied);
	  INCSTAT(lookups, frame->satisfied - old_satisfied);
	  frame->satisfied = old_satisfied;
	}

      assert (frame->satisfied >= first_learned_clause_idx);
      if (frame->satisfied == first_learned_clause_idx)
	break;

      /* Pesimistically assume that current clause is satisfied.  We may
       * need to fix this in case the clause turns out to be not satisfied.
       */
      assert (frame->satisfied > 0);
      frame->satisfied -= 1;

      clause = clauses[frame->satisfied];
      assert (!clause->external);	/* learned */

      assert (clause->iteration <= iterations);
      #if 0
      if (clause->iteration < iterations)
        break;
      #endif

      INCSTAT(lookups, 1);

      p = clause->literals;
      top_of_literals = p + size_literals (p);

      best = 0;
      best_lit = FALSE;
      tmp = FALSE;
      count = 0;

      while (p < top_of_literals)
	{
	  lit = *p++;
	  tmp = deref (lit);

	  if (tmp == TRUE)
	    break;

	  if (tmp == FALSE)
	    continue;

	  count++;

	  v = lit2var (lit);
	  if (best && cmp_var (best, v) >= 0)
	    continue;

	  best = v;
	  best_lit = lit;
	}

      if (tmp == TRUE)
	continue;

      frame->satisfied += 1;		/* fix it */

      if (best)
	{
	  if (count > 1)
	    return decide_phase (best);

	  assert (count == 1);
	  assert (!clause->connected);
	  assert (best_lit != FALSE);

	  trace1 ("<unit learned clause>");

	  reason = clause2reason (clause);
	  assign (best_lit, reason);

	  return TRUE;
	}

      assert (tmp == FALSE);

      reason = clause2reason (clause);
      *reason_ptr = reason;
      conflicts++;

      trace1 ("<empty learned clause>");
      trace2 ("conflict(%s)", reason2str (reason));

      return FALSE;
    }

  return decide_activity ();
}

/*------------------------------------------------------------------------*/

static Literal *
decide (Reason ** reason_ptr)
{
  Literal *res;

  check_invariant ();

  if (!iterations)
    iterations++;

  res = decide_connection (reason_ptr);

  if (is_two_valued_constant (res))
    return res;

  trace2 ("decide(%d)", lit2int (res));

  sum_levels += decision_level;
  decisions++;

  return res;
}

/*------------------------------------------------------------------------*/

static void
init_activities (void)
{
  Literal **q, **top_of_literals;
  Clause **p, *clause;

  for (p = clauses; p < top_of_clauses; p++)
    {
      clause = *p;
      q = clause->literals;
      top_of_literals = q + size_literals (q);
      while (q < top_of_literals)
	inc_variable_activity (lit2var (*q++));
    }

  log2 ("max logarithm variable activity %u", max_ld_activity);
}

/*------------------------------------------------------------------------*/

static void
postfix_order (Variable * v)
{
  Literal ** q, ** top_of_literals;
  Clause * clause;
  int sign_idx;
  Link *p;

  if (v->visited)
    return;

  v->visited = 1;

  for (sign_idx = 0; sign_idx <= 1; sign_idx++)
    {
      for (p = v->implications[sign_idx]; p; p = cdr(p))
	postfix_order (lit2var(car(p)));
    }

  for (sign_idx = 0; sign_idx <= 1; sign_idx++)
    {
      for (p = v->clauses[sign_idx]; p; p = cdr(p))
	{
	  clause = car(p);
	  q = clause->literals;
	  top_of_literals = q + size_literals (q);
	  while (q < top_of_literals)
	    postfix_order (lit2var(*q++));
	}
    }

  v->order = count_stack(order2var);
  push_stack (order2var,v);

  trace3 ("order(%d) = %d", var2idx (v), v->order);
}

/*------------------------------------------------------------------------*/

static unsigned
size_component (Component * c)
{
  assert (c->end > c->start);
  return c->end - c->start;
}

/*------------------------------------------------------------------------*/

static int
cmp_component (const void * p, const void * q)
{
  Component * c = (Component *) p;
  Component * d = (Component *) q;
  unsigned s = size_component (c);
  unsigned t = size_component (d);
  int res;

  if ((res = cmp_unsigned(s, t)))
    return res;

  assert (p == q || c->start != d->start);

  return cmp_unsigned(c->start, d->start);
}

/*------------------------------------------------------------------------*/

static void
init_order (void)
{
  Component component, *c;
  Variable * v, ** p;
  int size;

  v = variables + 1;
  if (v > top_of_variables)
    return;

  component.start = 0;

  for(;;v++)
    {
      if (v < top_of_variables && v->order)
	continue;

      component.end = count_stack (order2var);
      size = component.end - component.start;

      assert (size >= 0);

      if (size && 
	  (v == top_of_variables || 
	   (!component.start &&
	    !v->zombie &&
	    !is_top_level_assigned_variable (v))))
	{
	  push_stack (components, component);
	  component.start = component.end;
	}

      if (v == top_of_variables)
	break;

      postfix_order (v);
    }

  for (v = variables + 1; v < top_of_variables; v++)
    v->visited = 0;

  qsort (components, count_stack(components),
         sizeof(components[0]), cmp_component);

  for (c = components; c < top_of_components; c++)
    {
      log3 ("found component %u of size %d", 
	    c - components, size_component (c));

      for (p = order2var + c->start; p < order2var + c->end; p++)
	{
	  v = *p;
	  v->component = c;

	  trace4("component(%d) = %d of size %d",
		 var2idx (v), c - components, size_component (c));
	}
    }

  log2 ("found %u components", count_stack (components));
}

/*------------------------------------------------------------------------*/

static void
init_decide (void)
{
  find_zombies ();
  find_dependent ();
  find_independent ();
  init_order ();
  init_activities ();
}

/*------------------------------------------------------------------------*/

static void
push_frame (void)
{
  Frame *frame, * old_frame;

  frame = top_of_control++;
  frame->old_top_of_trail = top_of_trail;

  old_frame = frame - 1;

  if (frame == control)
    set_default_satisfied_interval (frame);
  else
    {
      frame->top_of_satisfied = old_frame->top_of_satisfied;
      frame->satisfied = old_frame->satisfied;
    }

  assert (decision_level + 1 == count_stack (control));
}

/*------------------------------------------------------------------------*/

static void
assume (Literal * lit)
{
  trace2 ("assume(%d)", lit2int (lit));
  decision_level++;
  push_frame ();
  assign (lit, DECISION_REASON);
}

/*------------------------------------------------------------------------*/

static size_t
get_bytes (void)
{
  size_t res;

  res = static_allocated_bytes + dynamic_allocated_bytes;

  return res;
}

/*------------------------------------------------------------------------*/

static double
get_mb (void)
{
  return get_bytes () / (double) (1 << 20);
}

/*------------------------------------------------------------------------*/
#if defined(LOG) || defined(STATS)
/*------------------------------------------------------------------------*/

static double
avg_level_as_double (void)
{
  if (!decisions)
    return 0;

  return sum_levels / (double) decisions;
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/
#ifdef LOG
/*------------------------------------------------------------------------*/

static void
report (void)
{
  double t;

  reports++;
  report_interval = NEW_REPORT_INTERVAL(report_interval);
  last_report_decisions = decisions;

  t = get_time ();

  log9 ("report(%.0f) %.1f %.1f %.0f %.1f %.1f %.1f %.1f",
	(double) reports, t, get_mb (), (double) conflicts,  
	avg_level_as_double (), avg_len_as_double (), 
	100.0 * learning_efficiency (),
	100.0 * connection_efficiency ());
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

static void
trigger_report (void)
{
#ifdef LOG
  int64 limit = last_report_decisions + report_interval;
  if (decisions >= limit)
    report ();
#endif
}

/*------------------------------------------------------------------------*/

static void
trigger_gc (void)
{
  unsigned limit;

  limit = last_gc_conflicts + GC_INTERVAL;
  if (conflicts >= limit)
    gc (0);
}

/*------------------------------------------------------------------------*/

static void
trigger_scale_down_variable_activities (void)
{
  unsigned limit;

  limit = last_scale_down_variable_activities_conflicts;
  limit += SCALE_DOWN_VARIABLE_ACTIVITIES_INTERVAL;
  if (conflicts < limit)
    return;

  last_scale_down_variable_activities_conflicts = conflicts;
  scale_down_variable_activities ();
}

/*------------------------------------------------------------------------*/

static void
trigger_scale_down_clause_activities (void)
{
  unsigned limit;

  limit = last_scale_down_clause_activities_conflicts;
  limit += SCALE_DOWN_CLAUSE_ACTIVIES_INTERVAL;
  if (conflicts < limit)
    return;

  last_scale_down_clause_activities_conflicts = conflicts;
  scale_down_clause_activities ();
}

/*------------------------------------------------------------------------*/

static void
restart (void)
{
  restarts++;

  log3 ("restart(%.0f) after %.0f conflicts",
        (double) restarts, (double) conflicts);

  undo (0);

  assert (unprocessed_clauses == top_of_trail);
  assert (unprocessed_implications == top_of_trail);
}

/*------------------------------------------------------------------------*/

static void
trigger_restart (void)
{
  unsigned limit;

  limit = last_restart_conflicts;
  limit += restart_interval;

  if (conflicts < limit)
    return;

  restart_interval += restart_interval / 2;

  last_restart_conflicts = conflicts;
  restart ();
}

/*------------------------------------------------------------------------*/

static void
tick (void)
{
  trigger_report ();
  trigger_gc ();
  trigger_scale_down_variable_activities ();
  trigger_scale_down_clause_activities ();
  trigger_restart ();
}

/*------------------------------------------------------------------------*/

static int
limit_reached (void)
{
  if (space_limit >= 0 && space_limit * (1 << 20) <= get_bytes ())
    {
      log2 ("space limit of %d MB reached", space_limit);
      state = SPACE_OUT_STATE;
      return 1;
    }

  if (decision_limit >= 0 && decision_limit >= decisions)
    {
      log2 ("decision limit of %.0f decisions reached", 
	    (double) decision_limit);

      state = DECISION_OUT_STATE;
      return 1;
    }

  if (time_limit >= 0 && time_spent_in_library () >= time_limit)
    {
      log2 ("time limit of %.1f seconds reached", time_limit);
      state = TIMED_OUT_STATE;
      return 1;
    }

  return 0;
}

/*------------------------------------------------------------------------*/

static Literal *
preprocess_units (void)
{
  if (!find_units ())
    return UNASSIGNED;

  if (assign_units () == FALSE)
    return FALSE;

  if (bcp ())
    return FALSE;

  gc (1);

  return UNASSIGNED;
}

/*------------------------------------------------------------------------*/

static void
preprocess_unates (void)
{
  int i;

  for (i = 0; i < MAX_FIND_UNATES; i++)
    {
      if (!find_unates ())
	break;

      assign_unates ();
      gc (1);
    }

  log2 ("preprocessing unates finished after %d iterations", i);
}

/*------------------------------------------------------------------------*/

static Literal *
preprocess (void)
{
  if (preprocess_units () == FALSE)
    return FALSE;

  preprocess_unates ();

  return UNASSIGNED;
}

/*------------------------------------------------------------------------*/

static Literal *
sat_aux (void)
{
  Reason *conflict;
  Literal *lit;

  if (preprocess () == FALSE)
    return FALSE;

  init_decide ();

  for (;;)
    {
      if (all_assigned ())
	return TRUE;

      if (limit_reached ())
	return UNASSIGNED;

      tick ();

      lit = decide (&conflict);
      if (lit == FALSE)
	goto ENTER_LOOP;

      if (lit != TRUE)
	assume (lit);

      while ((conflict = bcp ()))
      ENTER_LOOP:
	if (!backtrack (conflict))
	  return FALSE;
    }
}

/*------------------------------------------------------------------------*/

static Literal *
sat (void)
{
  Literal * res;

  res = sat_aux ();
#ifdef LOG
  report ();
#endif

  return res;
}
/*------------------------------------------------------------------------*/

#define BUFFERSZ 79

static char buffer[BUFFERSZ], *top_of_buffer;

/*------------------------------------------------------------------------*/

static void
flush_buffer (void)
{
  assert (top_of_buffer < buffer + BUFFERSZ);
  *top_of_buffer = 0;
  fputc ('v', out);
  fputs (buffer, out);
  fputc ('\n', out);
  top_of_buffer = buffer;
}

/*------------------------------------------------------------------------*/

static void
print_int (int d)
{
  char tmp[20];
  int len;

  sprintf (tmp, " %d", d);
  len = strlen (tmp);
  if (len + (top_of_buffer - buffer) >= BUFFERSZ)
    flush_buffer ();

  strcpy (top_of_buffer, tmp);
  top_of_buffer += len;
}

/*------------------------------------------------------------------------*/

static void
print_assigned_variable (Variable * v)
{
  int sign, idx;

  idx = var2idx (v);
  sign = lit2sign (v->assignment);

  print_int (sign * idx);
}

/*------------------------------------------------------------------------*/

static void
print_assignment (void)
{
  Variable *v;

  top_of_buffer = buffer;

  for (v = variables + 1; v < top_of_variables; v++)
    print_assigned_variable (v);

  print_int (0);

  if (top_of_buffer > buffer)
    flush_buffer ();
}

/*------------------------------------------------------------------------*/

static void
init_activity_queue (void)
{
  ActivityAnchor *anchor, **p;
  unsigned i;

  p = activity_queue;
  anchor = activity_anchors;

  for (i = 0; i < 32; i++)
    {
      anchor->first = anchor->last = 0;
      *p++ = anchor++;
    }
}

/*------------------------------------------------------------------------*/

static void
calculate_static_allocated_bytes (void)
{
  int ch, count;
  char name[40];
  FILE * file;
  size_t res;

  /* This is a rough guess.
   */
  static_allocated_bytes = 25 * (1 << 20) * sizeof (void*);

  sprintf (name, "/proc/%d/stat", getpid());
  file = fopen (name, "r");
  if (!file)
    return;

  count = 0;
  ch = fgetc (file);

  for(;;)
    {
      while (isspace (ch))
	ch = fgetc (file);

      if (ch == EOF)
	break;

      if (count++ == 22)
	{
	  res = 0;

	  while (isdigit (ch))
	    {
	      res = 10 * res + ch - '0';
	      ch = fgetc (file);
	    }

	  static_allocated_bytes = res;
	  break;
	}

      while (!isspace (ch))
	ch = fgetc (file);
    }

  fclose (file);
}

/*------------------------------------------------------------------------*/

static void
init (void)
{
  calculate_static_allocated_bytes ();

  top_of_trail = trail;
  unprocessed_clauses = trail;
  unprocessed_implications = trail;

  top_of_variables = variables;
  max_idx = 0;
  num_assigned = 0;
  num_zombies = 0;
  num_non_zombies = 0;
  num_independent = 0;
  num_dependent = 0;
  enable_cmp_var_independent = 0;

  init_stack (order2var);
  init_stack (components);
  active_component = UINT_MAX;

  init_stack (visited);

  init_activity_queue ();
  max_ld_activity = 0;
  ld_activity_scale = 0;
  max_ld_activity_scale = MAX_LD_ACTIVITY_SCALE;
  last_scale_down_variable_activities_conflicts = 0;

  top_of_control = control;
  decision_level = 0;
  push_frame ();
  sum_levels = 0;
  sum_length_learned_clauses = 0;
#ifdef SIG
  sum_length_learned_sigs = 0;
  sum_learned_sigs = 0;
#endif

  top_of_pushed_literals = pushed_literals;

  num_clauses = 0;
  num_empty_clauses = 0;
  num_binary_clauses = 0;
  num_implications = 0;
  num_learned_clauses = 0;
  valid_first_learned_clause_idx = 0;
  first_learned_clause_idx = 0;
  last_scale_down_clause_activities_conflicts = 0;
  init_stack (clauses);
#if defined(TRACE) || !defined(NDEBUG)
  next_clause_idx = 0;
#endif

  init_stack(units);
  old_count_units = 0;
  init_stack(unates);
  old_count_unates = 0;

  out = stdout;
  close_out = 0;
  in = stdin;
  name_in = "<stdin>";
  close_in = 0;

  chunks = 0;
  mem = 0;
  top_of_mem = 0;
  end_of_mem = 0;
  allocated_mem = 0;
  max_allocated_mem = 0;
  dynamic_allocated_bytes = 0;
  max_dynamic_allocated_bytes = 0;

  seconds = 0;

  decision_limit = -1;
  space_limit = -1;
  time_limit = -1;

  conflicts = 0;
  decisions = 0;
  iterations = 0;

  sum_learned_clauses = 0;
  sum_connected_clauses = 0;
  sum_multiple_used_learned_clauses = 0;
  sum_multiple_used_connected_clauses = 0;

  restarts = 0;
  restart_interval = RESTART_INTERVAL;
  last_restart_conflicts = 0;

  gcs = 0;
  dump_factor = DUMP_FACTOR;
  last_gc_conflicts = 0;

#ifdef LOG
  reports = 0;
  report_interval = REPORT_INTERVAL;
  last_report_decisions = 0;
#endif
#ifdef STATS

  adjustments = 0;
  adjusted_levels = 0;
  assignments = 0;
  backtracks = 0;
  collected = 0;
  decisions_based_on_activity = 0;
  hits = 0;
  implications = 0;
  irrelevant = 0;
  lookups = 0;
  skipped_backtracks = 0;
  top_level_assignments = 0;
  top_level_satisfied_clauses = 0;
  uips = 0;
  visits = 0;

#endif
}

/*------------------------------------------------------------------------*/

static void
release (void)
{
  if (close_out)
    fclose (out);

  out = 0;
  close_out = 0;

  if (close_in)
    fclose (in);

  in = stdin;
  close_in = 0;

  release_stack (components);
  release_stack (order2var);
  release_stack (visited);
  release_stack (units);
  release_stack (unates);
  release_stack (clauses);
  delete_chunks (chunks);

  assert (getenv ("NANOSAT_LEAK") || !allocated_mem);
  assert (getenv ("NANOSAT_LEAK") || !dynamic_allocated_bytes);
}

/*------------------------------------------------------------------------*/

#define USAGE \
"nanosat [<option> ...] [<input-file>]\n" \
"\n" \
"  -h | --help            print this command line summary\n" \
"  -d | --dump            print the parsed <input-file> in DIMACS format\n" \
"\n" \
"  -o <output-file>       set output and logging to <output-file>\n" \
"\n" \
"  --decision-limit=<n>   set decision limit to <n> decisions maximal\n" \
"  --time-limie=<t>       set time limit to <t> seconds\n" \
"  --space-limit=<s>      set space limit to <m> MB\n"

/*------------------------------------------------------------------------*/

static const char *
match (const char *pattern, const char *str)
{
  const char *p, *q;

  for (p = pattern, q = str; *p && *p == *q; p++, q++)
    ;

  return *p ? 0 : q;
}

/*------------------------------------------------------------------------*/

static int
is_int_str (const char *str)
{
  const char *p;

  p = str;
  if (*p == '-')
    p++;

  if (!*p)
    return 0;

  for (; *p; p++)
    if (!isdigit ((int) *p))
      return 0;

  return 1;
}

/*------------------------------------------------------------------------*/
/* -?[0-9][0-9]*\(\.[0-9][0-9]*\)
 */
static int
is_double_str (const char *str)
{
  const char *p;

  p = str;

  if (*p == '-')
    p++;

  if (!isdigit (*p))
    return 0;

  p++;

  while (isdigit (*p))
    p++;

  if (!*p)
    return 1;

  if (*p == '.')
    p++;

  if (!isdigit (*p))
    return 0;

  p++;
  while (isdigit (*p))
    p++;

  if (!*p)
    return 1;

  return 0;
}


/*------------------------------------------------------------------------*/
#ifdef TEST
/*------------------------------------------------------------------------*/

typedef void (*TestCase) (void);

/*------------------------------------------------------------------------*/

static void
whiteempty (void)
{
  /* do nothing */
}

/*------------------------------------------------------------------------*/

static void
whitealigned (void)
{
  fprintf (out, "(%d & (Word) variables) = %.0f\n",
           BIT_MASK, (double)(BIT_MASK & (Word) variables));
}

/*------------------------------------------------------------------------*/

static void
prldfloor (unsigned a)
{
  unsigned res;

  res = ldfloor (a);
  fprintf (out, "ldfloor (%u) = %u\n", a, res);
}

/*------------------------------------------------------------------------*/

static void
whiteldfloor (void)
{
  unsigned a;

  for (a = 0; a <= 35; a++)
    prldfloor (a);

  prldfloor (((unsigned) (1 << 31)) - 1);
  prldfloor (((unsigned) (1 << 31)) - 0);
  prldfloor (((unsigned) (1 << 31)) + 1);
  prldfloor (-1);
}

/*------------------------------------------------------------------------*/

static void
prldceil (unsigned a)
{
  unsigned res;

  res = ldceil (a);
  fprintf (out, "ldceil (%u) = %u\n", a, res);
}

/*------------------------------------------------------------------------*/

static void
whiteldceil (void)
{
  unsigned a;

  for (a = 0; a <= 35; a++)
    prldceil (a);

  prldceil (((unsigned) (1 << 31)) - 1);
  prldceil (((unsigned) (1 << 31)) - 0);
  prldceil (((unsigned) (1 << 31)) + 1);
  prldceil (-1);
}

/*------------------------------------------------------------------------*/

static int
activity_chain_length (ActivityAnchor * anchor)
{
  Variable *p;
  int res;

  res = 0;
  for (p = anchor->first; p; p = p->activity_chain.next)
    res++;

  return res;
}

/*------------------------------------------------------------------------*/

static void
dump_activity_queue (void)
{
  int i, len;

  fprintf (out, "max_ld_activity=%d\n", max_ld_activity);

  len = activity_chain_length (activity_queue[0]);
  fprintf (out, "activity_queue_length=%d", len);

  for (i = 1; i <= max_ld_activity; i++)
    {
      len = activity_chain_length (activity_queue[i]);
      fprintf (out, ",%d", len);
    }

  fputc ('\n', out);

}

/*------------------------------------------------------------------------*/

static void
dump_activities (void)
{
  ActivityAnchor *anchor;
  unsigned normalized;
  Variable *v;

  fprintf (out, "ld_activity_scale=%u\n", ld_activity_scale);

  dump_activity_queue ();

  for (v = variables + 1; v < top_of_variables; v++)
    {
      anchor = activity_queue[v->ld_activity];

      normalized = normalized_ld_activity (v);

      fprintf (out, "%d %u %u %u %u %u\n",
	       var2idx (v),
	       v->activity,
	       v->ld_activity,
	       v->ld_activity_scale,
	       normalized,
	       max_reverse_ldfloor (normalized));
    }
}

/*------------------------------------------------------------------------*/

static void
whiteactivity0 (void)
{
  int i;

  push_int_lit (1);
  push_int_lit (2);
  push_int_lit (3);
  push_int_lit (4);
  push_int_lit (5);
  push_int_lit (6);
  push_external_clause ();

  for (i = 0; i < (1 << 10); i++)
    inc_variable_activity (variables + 1);

  dump_activities ();
  for (i = 0; i < 6; i++)
    scale_down_variable_activities ();
  dump_activities ();

  for (i = 0; i < (1 << 10); i++)
    inc_variable_activity (variables + 2);

  inc_variable_activity (variables + 4);
  inc_variable_activity (variables + 5);
  inc_variable_activity (variables + 5);
  inc_variable_activity (variables + 5);
  inc_variable_activity (variables + 6);
  inc_variable_activity (variables + 6);
  inc_variable_activity (variables + 6);
  inc_variable_activity (variables + 6);

  dump_activities ();
  max_ld_activity_scale = ld_activity_scale;
  scale_down_variable_activities ();
  dump_activities ();

  for (i = 0; i < 7; i++)
    scale_down_variable_activities ();

  dump_activities ();
}

/*------------------------------------------------------------------------*/

static void
whitebacktrack0 (void)
{
  Reason *conflict;
  int res;

  push_int_lit (1);
  push_int_lit (2);
  push_external_clause ();

  push_int_lit (1);
  push_int_lit (-2);
  push_external_clause ();

  assert (count_stack (control) == 1);

  assume (var2lit (variables + 1, -1));
  conflict = bcp ();
  assert (conflict);
  assert (is_implication (conflict));
  res = backtrack (conflict);
  assert (res);
  assert (count_stack (control) == 1);
  assert (unprocessed_clauses == trail);
  assert (unprocessed_implications == trail);
  assert (top_of_trail == trail + 1);
  fprintf (out, "forced assignment %d\n", lit2int (trail[0]));
}

/*------------------------------------------------------------------------*/

static void
whitebacktrack1 (void)
{
  Reason *conflict;
  int res;

  push_int_lit (1);
  push_int_lit (-2);
  push_external_clause ();

  push_int_lit (1);
  push_int_lit (2);
  push_int_lit (-3);
  push_external_clause ();

  push_int_lit (1);
  push_int_lit (2);
  push_int_lit (3);
  push_external_clause ();

  assert (count_stack (control) == 1);

  assume (var2lit (variables + 1, -1));
  conflict = bcp ();
  assert (conflict);
  assert (!is_implication (conflict));
  res = backtrack (conflict);
  assert (res);
  assert (count_stack (control) == 1);
  assert (unprocessed_clauses == trail);
  assert (unprocessed_implications == trail);
  assert (top_of_trail == trail + 1);
  fprintf (out, "forced assignment %d\n", lit2int (trail[0]));
}

/*------------------------------------------------------------------------*/

static void
whiteindependent0 (void)
{
  push_int_lit (1);
  push_int_lit (2);
  push_external_clause ();
  push_int_lit (1);
  push_int_lit (3);
  push_external_clause ();
  push_int_lit (-1);
  push_int_lit (-2);
  push_int_lit (-3);
  push_external_clause ();

  find_dependent ();
  assert (num_dependent == 1);
  find_independent ();
  assert (num_independent == 2);
}

/*------------------------------------------------------------------------*/

static void
whiteindependent1 (void)
{
  push_int_lit (-1);
  push_int_lit (2);
  push_external_clause ();
  push_int_lit (-1);
  push_int_lit (3);
  push_external_clause ();
  push_int_lit (1);
  push_int_lit (-2);
  push_int_lit (-3);
  push_external_clause ();

  find_dependent ();
  assert (num_dependent == 1);
  find_independent ();
  assert (num_independent == 2);
}

/*------------------------------------------------------------------------*/

static void
whiteindependent2 (void)
{
  push_int_lit (-1);
  push_int_lit (2);
  push_external_clause ();
  push_int_lit (1);
  push_int_lit (-2);
  find_dependent ();

#if 0				/* DOES NOT WORK */
  find_dependent ();
  assert (num_dependent == 1);

  find_independent ();
  assert (num_independent == 1);
#endif
}

/*------------------------------------------------------------------------*/

static void
whitecomponent0 (void)
{
  push_int_lit (2);
  push_int_lit (4);
  push_external_clause ();
  push_int_lit (-2);
  push_int_lit (-4);
  push_external_clause ();
  push_int_lit (1);
  push_int_lit (3);
  push_int_lit (5);
  push_external_clause ();
  push_int_lit (-1);
  push_int_lit (-3);
  push_int_lit (-5);
  push_external_clause ();

  init_order ();

  assert (count_stack (components) == 2);
  assert (size_component(components + 0) == 2);
  assert (size_component(components + 1) == 3);
}

/*------------------------------------------------------------------------*/

static void
apiversion (void)
{
  fprintf (out, "%s\n", nanosat_version ());
}

/*------------------------------------------------------------------------*/

static void
apiid (void)
{
  fprintf (out, "c %s\n", nanosat_id ());
}

/*------------------------------------------------------------------------*/

static void
apispaceout (void)
{
  int res;
  res = 1;
  res &= nanosat_add (1);
  res &= nanosat_add (2);
  res &= nanosat_add (3);
  res &= nanosat_add (0);
  res &= nanosat_add (-1);
  res &= nanosat_add (-2);
  res &= nanosat_add (-3);
  res &= nanosat_add (0);
  assert (res);
  nanosat_set_space_limit (0);
  res = nanosat_sat ();
  assert (state == SPACE_OUT_STATE);
  assert (res == NANOSAT_SPACE_OUT);
}

/*------------------------------------------------------------------------*/

static void
apidecisionout (void)
{
  int res;
  res = 1;
  res &= nanosat_add (1);
  res &= nanosat_add (2);
  res &= nanosat_add (3);
  res &= nanosat_add (0);
  res &= nanosat_add (-1);
  res &= nanosat_add (-2);
  res &= nanosat_add (-3);
  res &= nanosat_add (0);
  assert (res);
  nanosat_set_decision_limit (0);
  res = nanosat_sat ();
  assert (state == DECISION_OUT_STATE);
  assert (res == NANOSAT_DECISION_OUT);
}

/*------------------------------------------------------------------------*/

static void
apisat0 (void)
{
  int res;
  res = 1;
  res &= nanosat_add (1);
  res &= nanosat_add (0);
  assert (res);
  res = nanosat_sat ();
  assert (state == DETERMINED_SATISFIABILITY_STATE);
  assert (res == NANOSAT_SATISFIABLE);
  res = nanosat_deref (1);
  assert (res == 1);
}

/*------------------------------------------------------------------------*/

static void
apisat1 (void)
{
  int res;
  res = 1;
  res &= nanosat_add (-1);
  res &= nanosat_add (0);
  assert (res);
  res = nanosat_sat ();
  assert (state == DETERMINED_SATISFIABILITY_STATE);
  assert (res == NANOSAT_SATISFIABLE);
  res = nanosat_deref (1);
  assert (res == 0);
}

/*------------------------------------------------------------------------*/

static int
phvar (int size, int i, int j)
{
  assert (size > 1);
  assert (1 <= i && i <= size);
  assert (1 <= j && j < size);

  return (i - 1) * size + j;
}

/*------------------------------------------------------------------------*/

static void
ph (int size)
{
  int p, q, h;

  assert (size > 1);

  for (p = 1; p <= size; p++)
    {
      for (h = 1; h < size; h++)
	nanosat_add (phvar (size, p, h));

      nanosat_add (0);
    }

  for (h = 1; h < size; h++)
    for (p = 1; p < size; p++)
      for (q = p + 1; q <= size; q++)
	{
	  nanosat_add (-phvar (size, p, h));
	  nanosat_add (-phvar (size, q, h));
	  nanosat_add (0);
	}
}

/*------------------------------------------------------------------------*/

static void
apipidgeon5 (void)
{
  int res;

  ph (5);
  res = nanosat_sat ();
  assert (res == NANOSAT_UNSATISFIABLE);
  nanosat_stats (out);
}

/*------------------------------------------------------------------------*/

static void
apitimeout (void)
{
  int res;
  ph (10);
  nanosat_set_time_limit (0.2);
  res = nanosat_sat ();
  assert (res == NANOSAT_TIME_OUT);
  nanosat_stats (out);
}

/*------------------------------------------------------------------------*/

static TestCase
find_test_case (const char *name)
{
#include "testcases.c"

  return 0;
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

static int
parse_int_opt (const char *pattern, const char *str,
	       int *opt_ptr, int *err_ptr)
{
  const char *arg;

  if (!(arg = match (pattern, str)))
    return 0;

  if (is_int_str (arg))
    {
      *opt_ptr = atoi (arg);
    }
  else if (arg[0])
    {
      fprintf (out, "*** '%s' argument is not an 'int'\n", pattern);
      *err_ptr = 1;
    }
  else
    {
      fprintf (out, "*** argument to '%s' is missing\n", pattern);
      *err_ptr = 1;
    }

  return 1;
}

/*------------------------------------------------------------------------*/

static int
parse_int64_opt (const char *pattern, const char *str,
	       int64 *opt_ptr, int *err_ptr)
{
  const char *arg, * p;
  int64 opt;
  int sign;

  if (!(arg = match (pattern, str)))
    return 0;

  if (is_int_str (arg))
    {
      opt = 0;
      p = arg;

      if (*p == '-')
	{
	  p++;
	  sign = -1;
	}
      else
	sign = 1;

      while (*p)
	opt = 10 * opt + (*p++ - '0');

      opt *= sign;
      *opt_ptr = opt;
    }
  else if (arg[0])
    {
      fprintf (out, "*** '%s' argument is not an 'int'\n", pattern);
      *err_ptr = 1;
    }
  else
    {
      fprintf (out, "*** argument to '%s' is missing\n", pattern);
      *err_ptr = 1;
    }

  return 1;
}

/*------------------------------------------------------------------------*/

static int
parse_double_opt (const char *pattern, const char *str,
		  double *opt_ptr, int *err_ptr)
{
  const char *arg;

  if (!(arg = match (pattern, str)))
    return 0;

  if (is_double_str (arg))
    {
      *opt_ptr = atof (arg);
    }
  else if (arg[0])
    {
      fprintf (out, "*** '%s' argument is not a 'double'\n", pattern);
      *err_ptr = 1;
    }
  else
    {
      fprintf (out, "*** argument to '%s' is missing\n", pattern);
      *err_ptr = 1;
    }

  return 1;
}

/*------------------------------------------------------------------------*/

static void
options (FILE * file)
{

  if (time_limit >= 0)
    print (file, "time limit %.1f seconds", time_limit);
  else
    print (file, "no time limit");

  if (decision_limit >= 0)
    print (file, "decision limit %.0f decisions", (double) decision_limit);
  else
    print (file, "no decision limit");

  if (space_limit >= 0)
    print (file, "space limit %d MB", space_limit);
  else
    print (file, "no space limit");

  print (file, "using %.1f MB statically allocated memory",
         static_allocated_bytes / (double)(1 << 20));
#ifdef LOG
  print (file, "garbage collection after %u conflicts", GC_INTERVAL);
  print (file, "scale down variables interval %d conflicts",
         SCALE_DOWN_VARIABLE_ACTIVITIES_INTERVAL);
  print (file, "scale down clauses interval %d conflicts",
         SCALE_DOWN_CLAUSE_ACTIVIES_INTERVAL);
#endif
}

/*------------------------------------------------------------------------*/
#ifdef STATS
/*------------------------------------------------------------------------*/

static double
hit_rate (void)
{
  if (!lookups)
    return 0;

  return hits / (double) lookups;
}

/*------------------------------------------------------------------------*/
#endif
/*------------------------------------------------------------------------*/

static void
stats (FILE * file)
{
  double t = time_spent_in_library ();
  double mb = get_mb ();
#ifdef STATS
  print (file, "%.0f gcs (%.0f collected, %.0f irrelevant, %.0f satisfied)",
         (double) gcs, (double) collected, 
	 (double) irrelevant, (double) top_level_satisfied_clauses);

  print (file, "%.0f restarts", (double) restarts);
  print (file, "%.0f conflicts (%.1f average length)", 
         (double) conflicts, avg_len_as_double ());
#ifdef SIG
  print (file, "%.0f learned signatures (%.1f average length)",
         (double) sum_learned_sigs,
	 sum_learned_sigs ? 
	   sum_length_learned_sigs / (double) sum_learned_sigs : 0.0);
#endif
  print (file, "%.0f decisions (%.0f based on activity, %.1f average level)",
         (double) decisions, 
	 (double) decisions_based_on_activity,
	 avg_level_as_double());
  print (file, "%.1f%% learning efficiency (%.1f%% for connected)",
         100.0 * learning_efficiency(), 100.0 * connection_efficiency());
  print (file, "%.0f lookups (%.0f hits = %.1f%%)",
         (double) lookups, (double) hits, 100.0 * hit_rate ());
  print (file, "%.0f backtracks (%.0f skipped, %.0f uips)", 
         (double) backtracks, (double) skipped_backtracks, (double) uips);
  print (file, "%.0f assignments (%.0f implications, %.0f visits)",
         (double) assignments, (double) implications, (double) visits);
  print (file, "%u iterations (%.0f top level assignments)", 
         iterations, (double) top_level_assignments);
  print (file, "%.0f adjustments (%.0f levels)", 
         (double) adjustments, (double) adjusted_levels);
#else
  print (file, "%.0f conflicts", (double) conflicts);
  print (file, "%.0f decisions", (double) decisions);
#endif
  print (file, "%.1f seconds", t);
  print (file, "%.1f MB", mb);
}

/*------------------------------------------------------------------------*/

const char *
id (void)
{
  return "$Id: nanosat.c,v 1.181 2005/09/07 09:33:38 biere Exp $";
}

/*------------------------------------------------------------------------*/

static const char *
version (void)
{
  return VERSION;
}

/*------------------------------------------------------------------------*/

static void
enter (void)
{
  if (!entered++)
    entered_time = get_time ();
}

/*------------------------------------------------------------------------*/

static void
leave (void)
{
  assert (entered > 0);

  if (!--entered)
    seconds += delta_time ();
}

/*------------------------------------------------------------------------*/
/**************************** EXTERNAL PART *******************************/
/*------------------------------------------------------------------------*/

#define external_assert(f,b) \
do \
  { \
    if (!(b)) \
      { \
	perror ("libnanosat: invalid API usage in " f "(): " # b); \
	abort(); \
      } \
  } \
while(0)

/*------------------------------------------------------------------------*/

int
nanosat_main (int argc, char **argv)
{
  int err, done, i, dump_only;
  int external_res;
  FILE *new_out;
  Literal *res;
#ifdef TEST
  const char *arg;
#endif

  external_assert ("nanosat_main", state == UNINITIALIZED_STATE);
  state = INITIALIZED_STATE;
  
  assert (!entered);
  enter ();
  init ();

  done = err = 0;
  dump_only = 0;

  for (i = 1; !err && !done && i < argc; i++)
    {
      if (!strcmp (argv[i], "-h") || !strcmp (argv[i], "--help"))
	{
	  fputs (USAGE, out);
	  done = 1;
	}
#ifdef TEST
      else if ((arg = match ("--white", argv[i])) ||
               (arg = match ("--api", argv[i])))
	{
	  TestCase test_case = find_test_case (arg);
	  if (test_case)
	    {
	      test_case ();
	      done = 1;
	    }
	  else
	    {
	      fprintf (out, "*** '%s' invalid test case name\n", argv[i]);
	      err = 1;
	    }
	}
#endif
      else if (parse_int64_opt ("--decision-limit=", argv[i],
			      &decision_limit, &err))
	{
	  /* do nothing */
	}
      else if (parse_int_opt ("--space-limit=", argv[i], &space_limit, &err))
	{
	  /* do nothing */
	}
      else if (parse_double_opt ("--time-limit=", argv[i], &time_limit, &err))
	{
	  /* do nothing */
	}
      else if (!strcmp (argv[i], "-o"))
	{
	  if (i + 1 == argc)
	    {
	      fprintf (out, "*** argument to '-o' missing\n");
	      err = 1;
	    }
	  else if (!(new_out = fopen (argv[++i], "w")))
	    {
	      fprintf (out, "*** could not write '%s'\n", argv[i]);
	      err = 1;
	    }
	  else
	    {
	      if (close_out)
		fclose (out);

	      out = new_out;
	      close_out = 1;
	    }
	}
      else if (!strcmp (argv[i], "-d"))
	{
	  dump_only = 1;
	}
      else if (argv[i][0] == '-')
	{
	  fprintf (out, "*** invalid option '%s'\n", argv[i]);
	  err = 1;
	}
      else if (close_in)
	{
	  fprintf (out, "*** multiple files\n");
	  err = 1;
	}
      else if (!(in = fopen (argv[i], "r")))
	{
	  fprintf (out, "*** could not open '%s'\n", argv[i]);
	  err = 1;
	}
      else
	{
	  close_in = 1;
	  name_in = argv[i];
	}
    }

  res = UNASSIGNED;

  if (!done && !err)
    {
      if (!dump_only)
	{
	  print (out, "NanoSat Sat Solver Version %s", version());
	  print (out, "(C)opyright 2003-2004 by Armin Biere");
	  print (out, id ());
	  options (out);
	}

      if (parse ())
	{
	  log3 ("p cnf %u %u", max_idx, num_clauses);

	  if (dump_only)
	    dump (out, 1);
	  else
	    {
	      res = sat ();

	      if (res == FALSE)
		fprintf (out, "s UNSATISFIABLE\n");
	      else if (res == TRUE)
		{
		  fprintf (out, "s SATISFIABLE\n");
		  print_assignment ();
		}
	      else
		fprintf (out, "s UNKNOWN\n");
	    }
	}
      else
	err = 1;
    }

  if (!done && !err)
    stats (out);

  release ();

  if (err)
    external_res = NANOSAT_ERROR;
  else if (done)
    external_res = NANOSAT_EXIT;
  else if (res == UNASSIGNED)
    {
      if (state == TIMED_OUT_STATE)
	external_res = NANOSAT_TIME_OUT;
      else if (state == DECISION_OUT_STATE)
	external_res = NANOSAT_DECISION_OUT;
      else if (state == SPACE_OUT_STATE)
	external_res = NANOSAT_SPACE_OUT;
      else
	external_res = NANOSAT_EXIT;
    }
  else if (res == TRUE)
    {
      state = DETERMINED_SATISFIABILITY_STATE;
      external_res = NANOSAT_SATISFIABLE;
    }
  else
    {
      state = DETERMINED_UNSATISFIABILITY_STATE;
      external_res = NANOSAT_UNSATISFIABLE;
    }

  leave ();
  assert (!entered);
  state = UNINITIALIZED_STATE;

  return external_res;
}

/*------------------------------------------------------------------------*/

void
nanosat_init (void)
{
  external_assert ("nanosat_init", state == UNINITIALIZED_STATE);
  state = INITIALIZED_STATE;

  assert (!entered);
  enter ();
  init ();
  leave ();
}

/*------------------------------------------------------------------------*/

void
nanosat_release (void)
{
  external_assert ("nanosat_release", state != UNINITIALIZED_STATE);
  state = UNINITIALIZED_STATE;

  assert (!entered);
  enter ();
  release ();
  leave ();
}

/*------------------------------------------------------------------------*/

void
nanosat_set_output_file (FILE * file)
{
  external_assert ("nanosat_set_output_file", state == INITIALIZED_STATE);
  assert (!close_out);
  out = file;
}

/*------------------------------------------------------------------------*/

int 
nanosat_add (int lit)
{
  int res, idx;

  enter ();

  idx = int2nat (lit);

  if (idx < MAXVARS && !limit_reached ())
    {
      res = 1;

      if (lit)
	push_int_lit (lit);
      else
	push_external_clause ();
    }
  else
    res = 0;

  leave ();

  return res;
}

/*------------------------------------------------------------------------*/

void
nanosat_dump (FILE * file)
{
  dump (file, 1);
}

/*------------------------------------------------------------------------*/

void
nanosat_set_space_limit (int mb)
{
  external_assert ("nanosat_set_space_limit", state == INITIALIZED_STATE);
  enter ();
  space_limit = mb;
  leave ();
}

/*------------------------------------------------------------------------*/

void
nanosat_set_time_limit (double seconds)
{
  external_assert ("nanosat_set_time_limit", state == INITIALIZED_STATE);
  enter ();
  time_limit = seconds;
  leave ();
}

/*------------------------------------------------------------------------*/

void
nanosat_set_decision_limit (int limit)
{
  external_assert ("nanosat_set_decision_limit", state == INITIALIZED_STATE);
  enter ();
  decision_limit = decisions;
  leave ();
}

/*------------------------------------------------------------------------*/

int
nanosat_sat (void)
{
  int external_res;
  Literal * res;

  external_assert ("nanosat_sat", state == INITIALIZED_STATE);
  enter ();

  res = sat ();

  if (res == TRUE)
    {
      state = DETERMINED_SATISFIABILITY_STATE;
      external_res = NANOSAT_SATISFIABLE;
    }
  else if (res == FALSE)
    {
      state = DETERMINED_UNSATISFIABILITY_STATE;
      external_res = NANOSAT_UNSATISFIABLE;
    }
  else if (state == TIMED_OUT_STATE)
    external_res = NANOSAT_TIME_OUT;
  else if (state == DECISION_OUT_STATE)
    external_res = NANOSAT_DECISION_OUT;
  else
    {
      assert (state == SPACE_OUT_STATE);
      external_res = NANOSAT_SPACE_OUT;
    }
  leave ();

  return external_res;
}

/*------------------------------------------------------------------------*/

int
nanosat_deref (int lit)
{
  int res;

  external_assert ("nanosat_deref", state == DETERMINED_SATISFIABILITY_STATE);
  enter ();
  res = deref (int2lit (lit)) == TRUE;
  leave ();

  return res;
}

/*------------------------------------------------------------------------*/

void
nanosat_options (FILE * file)
{
  enter ();
  options (file);
  leave ();
}

/*------------------------------------------------------------------------*/

void
nanosat_stats (FILE * file)
{
  enter ();
  stats (file);
  leave ();
}

/*------------------------------------------------------------------------*/

const char *
nanosat_id (void)
{
  const char * res;
  enter ();
  res = id ();
  leave ();
  return res;
}

/*------------------------------------------------------------------------*/

const char *
nanosat_version (void)
{
  const char * res;
  enter ();
  res = version ();
  leave ();
  return res;
}

/*------------------------------------------------------------------------*/

double
nanosat_seconds (void)
{
  return time_spent_in_library ();
}

/*------------------------------------------------------------------------*/

size_t
nanosat_bytes (void)
{
  return get_bytes ();
}
