#ifndef NANOSAT_INCLUDED
#define NANOSAT_INCLUDED

/*------------------------------------------------------------------------*/

#define NANOSAT_PREFIX "c nst "

/*------------------------------------------------------------------------*/

#define NANOSAT_SATISFIABLE 10
#define NANOSAT_UNSATISFIABLE 20
#define NANOSAT_EXIT 0
#define NANOSAT_ERROR 1
#define NANOSAT_TIME_OUT 2
#define NANOSAT_SPACE_OUT 3
#define NANOSAT_DECISION_OUT 4

/*------------------------------------------------------------------------*/
/* This is the application interface, which is mainly used for attaching the
 * black box test suite.  The result values are those defined above.  The
 * library is left in a well defined state after 'nanosat_main' finishes.
 * Even if an error condition arises, 'nanosat_main' can be called again.
 */
int nanosat_main (int argc, char **argv);

/*------------------------------------------------------------------------*/
/* Beside the application interface, we have a low level interface, based on
 * the remaining functions declared below.  If the rules specified below are
 * followed, then this second API is also reentrant, but not thread-safe.
 *
 * First you need to initialize the library with 'nanosat_init' before you
 * can use any of the other external functions below.  It is also necessary
 * to release the library with 'nanosat_release', if the library is used
 * again from the same executable.
 *
 * The effect of calling 'nanosat_init' twice in a row without calling
 * 'nanosat_reset' is undefined.  It is also undefined to call
 * 'nanosat_release' without a previous call to 'nanosat_init' or calling it
 * twice in a row.
 */
void nanosat_init (void);
void nanosat_release (void);

/*------------------------------------------------------------------------*/
/* Verbose output is enabled at compile time by adding the '--log' or even
 * the '--trace' flag to './configure'.  This will produce verbose output to 
 * a log file, which can be specified by 'nanosat_set_output_file'.  This
 * log file is also used for the 'nanosat_{options,version,id,stats}'
 * functions below.  Default is <stdout>.
 */
void nanosat_set_output_file (FILE *);

/*------------------------------------------------------------------------*/
/* To add clauses you can use 'nanosat_add' by calling it with each literal
 * of the clause followed by 'nanosat_add(0)'.  The return value of
 * 'nanosat_add' is zero if the literal is invalid.  A literal is invalid if
 * it is zero or larger than the maximal number of variables supported.  If
 * one of the resource limits is reached, zero is returned as well.
 */
int nanosat_add (int literal);

/*------------------------------------------------------------------------*/
/* Dump all clauses in DIMACS format.
 */
void nanosat_dump (FILE * file);

/*------------------------------------------------------------------------*/
/* Set resource limits with these functions.  A negative value disables
 * the resource limit.  You need to call 'funex_init' first.
 */
void nanosat_set_time_limit (double seconds);
void nanosat_set_space_limit (int mb);
void nanosat_set_decision_limit (int decisions);

/*------------------------------------------------------------------------*/
/* Will return 'SATISFIABLE', 'UNSATISFIABLE', 'TIME_OUT', 'SPACE_OUT', or
 * 'DECISION_OUT', as defined above.
 */
int nanosat_sat ();

/*------------------------------------------------------------------------*/
/* With 'nanosat_deref' you can retrieve the satisfying assignment for the
 * current instance after 'nanosat_sat' was called and returned
 * 'NANOSAT_SATISFIABLE'.  The result is either '0' or '1'.  If the given
 * literal is invalid, see 'nanosat_add', then zero is returned.
 */
int nanosat_deref (int literal);

/*------------------------------------------------------------------------*/
/* Print run time options or statistics to the specified file.
 */
void nanosat_options (FILE *);
void nanosat_stats (FILE *);

/*------------------------------------------------------------------------*/

const char * nanosat_version (void);
const char * nanosat_id (void);

/*------------------------------------------------------------------------*/

double nanosat_seconds ();
size_t nanosat_bytes ();

/*------------------------------------------------------------------------*/

#endif
