/*------------------------------------------------------------------------*/

#include <assert.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>

/*------------------------------------------------------------------------*/

#include "config.h"
#include "nanosat.h"

/*------------------------------------------------------------------------*/

typedef struct TestCase TestCase;
typedef struct TestSuite TestSuite;

/*------------------------------------------------------------------------*/
/* We only use some kind of black box test cases.  The 'argc' and 'argv' are
 * read from 'log/options' as space separated strings.  Usually there is one
 * set of arguments listed on one line.  However the line may be continued
 * by starting the next line with a space character.
 */
struct TestCase
{
  int argc;
  char **argv;
  int res;
};

/*------------------------------------------------------------------------*/

struct TestSuite
{
  int argc;			/* original 'argc' */
  char **argv;			/* original 'argv' */

  int test_cases_count;		/* number of parsed test cases */
  int test_cases_size;		/* size of test case array */
  TestCase *test_cases;		/* test case array */

  int count;			/* number of matched test cases */
  int ok;			/* number of successful test cases */
  int failed;			/* number of failed test cases */
  int skipped;			/* number of skipped test cases */

  char *buffer;			/* buffer for reading the tokens */
  int buffer_size;
  int buffer_count;

  int erase_line;
  int check;

  FILE *file;
};

/*------------------------------------------------------------------------*/

static char *
my_strdup (const char *str)
{
  char *res;

  res = malloc (strlen (str) + 1);
  strcpy (res, str);

  return res;
}

/*------------------------------------------------------------------------*/

static int
is_dumb_terminal (void)
{
  const char *str;

  str = getenv ("TERM");

  if (!str)
    return 1;

  if (!strcmp (str, "emacs"))
    return 1;

  if (!strcmp (str, "dumb"))
    return 1;

  return 0;
}

/*------------------------------------------------------------------------*/

#define USAGE \
"usage: test [<option> ...] [<pattern> ...]\n" \
"\n" \
"where <option> is one of the following:\n" \
"\n" \
"  -h | --help        print this command line option summary and exit\n" \
"  -k | --keep        keep all results (default for non tty output)\n" \
"  --check[=<level>]  enable expensive checking code\n"

/*------------------------------------------------------------------------*/

static const char *
match (const char *pattern, const char *str)
{
  const char *p, *q;

  for (p = str, q = pattern; *q && (*p == *q); p++, q++)
    ;

  return *q ? 0 : p;
}

/*------------------------------------------------------------------------*/

static TestSuite *
new_TestSuite (int argc, char **argv)
{
  const char *arg;
  TestSuite *res;
  int i;

  res = (TestSuite *) malloc (sizeof (*res));
  memset (res, 0, sizeof (*res));

  res->argc = argc;
  res->argv = argv;

  res->test_cases_count = 0;
  res->test_cases_size = 1000;
  res->test_cases = (TestCase *)
    calloc (res->test_cases_size, sizeof (TestCase));

  res->buffer_size = 1;
  res->buffer = (char *) malloc (res->buffer_size);

  res->erase_line = 1;

  for (i = 1; i < argc; i++)
    {
      if (!strcmp (argv[i], "-h") || !strcmp (argv[i], "--help"))
	{
	  printf (USAGE);
	  exit (0);
	}
      else if (!strcmp (argv[i], "-k") || !strcmp (argv[i], "--keep"))
	{
	  res->erase_line = 0;
	}
      else if (!strcmp (argv[i], "--check"))
	{
	  res->check++;
	}
      else if ((arg = match ("--check=", argv[i])))
	{
	  res->check = atoi (arg);
	}
      else if (argv[i][0] == '-')
	{
	  fprintf (stderr,
		   "*** test: invalid option '%s' (try '-h')\n", argv[i]);
	  exit (1);
	}
    }

  if (!isatty (1) || is_dumb_terminal ())
    res->erase_line = 0;

  return res;
}

/*------------------------------------------------------------------------*/

static char *
check_arg (int check)
{
  char buffer[30];
  sprintf (buffer, "--check=%d", check);
  return my_strdup (buffer);
}

/*------------------------------------------------------------------------*/

static TestCase *
read_TestCase (TestSuite * ts)
{
  int argc, size, ch, count_non_space_chars;
  char **argv, *token;
  TestCase *tc;
  argc = 0;


RESTART_PARSING_LINES:

  ts->buffer_count = 0;
  count_non_space_chars = 0;

  for (;;)
    {
      ch = fgetc (ts->file);

    REENTER_LOOP_WITHOUT_READING_CHAR:

      if (ch == EOF)
	break;

      if (ch == '\n')
	{
	  ch = fgetc (ts->file);
	  if (ch != ' ' && ch != '\t')
	    {
	      ungetc (ch, ts->file);
	      break;
	    }
	}
      else if (ch == '#')
	{
	  while ((ch = fgetc (ts->file)) != '\n' && ch != EOF)
	    ;

	  goto REENTER_LOOP_WITHOUT_READING_CHAR;
	}

      if (ts->buffer_count + 1 >= ts->buffer_size)
	{
	  ts->buffer_size *= 2;
	  ts->buffer = (char *) realloc (ts->buffer, ts->buffer_size);
	}

      ts->buffer[ts->buffer_count++] = ch;
      count_non_space_chars += !isspace (ch);
    }

  if (ch == EOF && !ts->buffer_count)
    return 0;

  if (!count_non_space_chars)
    goto RESTART_PARSING_LINES;

  assert (ts->buffer_count < ts->buffer_size);
  ts->buffer[ts->buffer_count] = 0;

  size = 1;
  argc = 0;
  argv = (char **) malloc (size * sizeof (char *));

  argv[argc++] = my_strdup (strtok (ts->buffer, " \t"));

  if (ts->check > 0)
    {
      if (argc >= size)
	{
	  size *= 2;
	  argv = (char **) realloc (argv, size * sizeof (char *));
	}

      argv[argc++] = check_arg (ts->check);
    }

  while ((token = strtok (0, " \t")))
    {
      if (argc >= size)
	{
	  size *= 2;
	  argv = (char **) realloc (argv, size * sizeof (char *));
	}

      argv[argc++] = my_strdup (token);
    }

  argv = (char **) realloc (argv, argc * sizeof (char *));

  if (ts->test_cases_count >= ts->test_cases_size)
    {
      ts->test_cases_size *= 2;
      ts->test_cases = (TestCase *)
	realloc (ts->test_cases, ts->test_cases_size * sizeof (TestCase));
    }

  tc = &ts->test_cases[ts->test_cases_count++];
  tc->argc = argc;
  tc->argv = argv;
  tc->res = -1;

  return tc;
}

/*------------------------------------------------------------------------*/

static void
read_TestSuite (TestSuite * ts)
{
  if (!(ts->file = fopen ("log/options", "r")))
    {
      fprintf (stderr, "*** test: could not read 'log/options'\n");
      exit (1);
    }

  while (read_TestCase (ts))
    ;
}

/*------------------------------------------------------------------------*/

static int
match_TestCase (TestSuite * ts, TestCase * tc)
{
  int num_pattern, found, i;

  num_pattern = 0;
  found = 0;

  for (i = 1; !found && i < ts->argc; i++)
    {
      if (ts->argv[i][0] == '-')
	continue;

      num_pattern++;
      found = (match (ts->argv[i], tc->argv[0]) != 0);
    }

  if (!num_pattern)
    found = 1;

  return found;
}

/*------------------------------------------------------------------------*/

static int
next_char_in_log_file (FILE * file, int last_char)
{
  int ch;

  ch = fgetc (file);

  while (ch == 'c' && last_char == '\n')
    {
      while ((ch = fgetc (file)) != EOF && ch != '\n')
	;

      if (ch != EOF)
	{
	  last_char = '\n';
	  ch = fgetc (file);
	}
    }

  return ch;
}

/*------------------------------------------------------------------------*/
/* Compare two files for differences.  The result is '1' iff no differences
 * were encountered and both files could be accessed.  If one file could not
 * be openend, then '0' is returned.
 */
static int
cmp_files (char *name0, char *name1)
{
  int res, ch[2], last_char[2], done;
  FILE *f[2];

  last_char[0] = last_char[1] = '\n';

  f[0] = fopen (name0, "r");
  if (f[0])
    {
      f[1] = fopen (name1, "r");
      if (f[1])
	{
	  do
	    {
	      ch[0] = next_char_in_log_file (f[0], last_char[0]);
	      ch[1] = next_char_in_log_file (f[1], last_char[1]);
	      res = (ch[0] == ch[1]);
	      last_char[0] = ch[0];
	      last_char[1] = ch[1];
	      done = (ch[0] == EOF);
	    }
	  while (res && !done);
	  fclose (f[1]);
	}
      else
	res = 0;
      fclose (f[0]);
    }
  else
    res = 0;

  return res;
}

/*------------------------------------------------------------------------*/

static int
run_TestCase (TestSuite * ts, TestCase * tc)
{
  char *out_name, *log_name, *name;
  int i, my_argc, res;
  char **my_argv;

  name = tc->argv[0];
  log_name = (char *) malloc (strlen (name) + 20);
  out_name = (char *) malloc (strlen (name) + 20);
  sprintf (log_name, "log/%s.log", name);
  sprintf (out_name, "log/%s.out", name);

  my_argc = tc->argc + 2;
  my_argv = (char **) malloc (sizeof (char *) * my_argc);
  my_argv[0] = tc->argv[0];
  my_argv[1] = "-o";
  my_argv[2] = log_name;
  for (i = 1; i < tc->argc; i++)
    my_argv[i + 2] = tc->argv[i];

  (void) nanosat_main (my_argc, my_argv);

  res = cmp_files (log_name, out_name);

  free (log_name);
  free (out_name);
  free (my_argv);

  return res;
}

/*------------------------------------------------------------------------*/

static int
print_str (const char *str, int old_len)
{
  int res;

  res = old_len + strlen (str);
  fputs (str, stdout);

  return res;
}

/*------------------------------------------------------------------------*/

static void
print_chars (int ch, int count)
{
  int i;

  for (i = 0; i < count; i++)
    fputc (ch, stdout);
}

/*------------------------------------------------------------------------*/

static void
run_TestSuite (TestSuite * ts)
{
  int i, res, len;
  TestCase *tc;

  for (i = 0; i < ts->test_cases_count; i++)
    {
      tc = ts->test_cases + i;
      if (!match_TestCase (ts, tc))
	continue;

#ifndef TEST
      if (match ("white", tc->argv[0]) || match ("api", tc->argv[0]))
	{
	  ts->skipped++;
	  continue;
	}
#endif

      ts->count++;
      len = print_str (tc->argv[0], 0);
      len = print_str (" ...", len);
      fflush (stdout);
      res = run_TestCase (ts, tc);
      tc->res = res;
      if (res)
	{
	  len = print_str (" ok", len);
	  ts->ok++;
	}
      else
	{
	  len = print_str (" failed", len);
	  ts->failed++;
	}

      if (res && ts->erase_line)
	{
	  print_chars ('', len);
	  print_chars (' ', len);
	  print_chars ('', len);
	  fflush (stdout);
	}
      else
	fputc ('\n', stdout);
    }
}

/*------------------------------------------------------------------------*/

static void
report_TestSuite (TestSuite * ts)
{
  printf ("%d count, %d ok, %d failed, %d skipped.\n",
	  ts->test_cases_count, ts->ok, ts->failed, ts->skipped);
}

/*------------------------------------------------------------------------*/

static void
release_TestCase (TestCase * tc)
{
  int i;

  for (i = 0; i < tc->argc; i++)
    free (tc->argv[i]);

  free (tc->argv);
}

/*------------------------------------------------------------------------*/

static void
delete_TestSuite (TestSuite * ts)
{
  TestCase *tc;
  int i;

  if (ts->file)
    fclose (ts->file);

  free (ts->buffer);

  for (i = 0; i < ts->test_cases_count; i++)
    {
      tc = ts->test_cases + i;
      release_TestCase (tc);
    }

  free (ts->test_cases);
  free (ts);
}

/*------------------------------------------------------------------------*/

int
main (int argc, char **argv)
{
  TestSuite *ts;

  ts = new_TestSuite (argc, argv);
  read_TestSuite (ts);
  run_TestSuite (ts);
  report_TestSuite (ts);
  delete_TestSuite (ts);

  return 0;
}
